/**
  *  \file client/dialogs/buildbilldlg.cc
  *  \brief Starbase Ship Build Detailed Bill
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/buildbilldlg.h"
#include "client/widgets/billdisplay.h"
#include "game/actions/baseship.h"
#include "game/planet.h"
#include "ui/layout.h"
#include "ui/ui.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/window.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/planetform.h"

namespace {
    /** Implementation of GStarbaseBuildExecuter that generates a detailed bill. */
    class BillingActor : public GStarbaseBuildExecuter {
        const GPlanet& planet;
        WBill& bill;
     public:
        BillingActor(const GPlanet& planet, WBill& bill)
            : planet(planet), bill(bill)
            { }
        void setTech(TTechLevel area, int value);
        void setHullSlot(int index, int value, int free);
        void setEngine(int index, int value, int free);
        void setBeam(int index, int value, int free);
        void setLauncher(int index, int value, int free);
        void accountHull(int number, int count, int free);
    };
}

void
BillingActor::setTech(TTechLevel area, int value)
{
    if (planet.getBaseTechLevel(area) != value) {
        static const char*const names[] = {
            N_("Engine tech upgrade"),
            N_("Hull tech upgrade"),
            N_("Beam tech upgrade"),
            N_("Torpedo tech upgrade")
        };
        GCost cost;
        cost.set(el_Money, getBaseTechCost(planet.getOwner(), planet.getBaseTechLevel(area), value));
        bill.push_back(new WBillItem(0, _(names[area]), cost, value - planet.getBaseTechLevel(area)));
    }
}

void
BillingActor::setHullSlot(int index, int value, int free)
{
    int hull_id = getTruehull(planet.getOwner(), index);
    int diff = value - planet.getBaseHullStoreSlot(index);
    if (diff) {
        bill.push_back(new WBillItem(0, getHull(hull_id).getName(), getHull(hull_id).getCost() * diff, diff));
    }
    if (free) {
        bill.push_back(new WBillItem(0, format(_("From storage: %s"), getHull(hull_id).getName()), GCost(), free));
    }
}

void
BillingActor::setEngine(int index, int value, int free)
{
    int diff = value - planet.getBaseEngineStore(index);
    if (diff) {
        bill.push_back(new WBillItem(0, getEngine(index).getName(), getEngine(index).getCost() * diff, diff));
    }
    if (free) {
        bill.push_back(new WBillItem(0, format(_("From storage: %s"), getEngine(index).getName()), GCost(), free));
    }
}

void
BillingActor::setBeam(int index, int value, int free)
{
    int diff = value - planet.getBaseBeamStore(index);
    if (diff) {
        bill.push_back(new WBillItem(0, getBeam(index).getName(), getBeam(index).getCost() * diff, diff));
    }
    if (free) {
        bill.push_back(new WBillItem(0, format(_("From storage: %s"), getBeam(index).getName()), GCost(), free));
    }
}

void
BillingActor::setLauncher(int index, int value, int free)
{
    int diff = value - planet.getBaseLauncherStore(index);
    if (diff) {
        bill.push_back(new WBillItem(0, getTorp(index).getName(), getTorp(index).getCost() * diff, diff));
    }
    if (free) {
        bill.push_back(new WBillItem(0, format(_("From storage: %s"), getTorp(index).getName()), GCost(), free));
    }
}

void
BillingActor::accountHull(int number, int count, int free)
{
    if (count) {
        bill.push_back(new WBillItem(0, getHull(number).getName(), getHull(number).getCost() * count, count));
    }
    if (free) {
        bill.push_back(new WBillItem(0, format(_("From storage: %s"), getHull(number).getName()), GCost(), free));
    }
}


/** Detailed bill dialog.
    \param planet Planet who pays the build order
    \param action Action that contains the build order */
void
doBuildDetailedBill(const GPlanet& planet, GStarbaseBuildShipAction& action)
{
    // Generate the bill
    WBill bill;
    BillingActor actor(planet, bill);
    action.perform(actor);

    // For convenience, also include fighter bays
    if (int bays = getHull(action.getBuildOrder().hull).getNumBays())
        bill.push_back(new WBillItem(0, _("Fighter Bay"), GCost(), bays));

    // Build a dialog
    //   VBox
    //     VBox
    //       WBillDisplay
    //       WBillTotalDisplay
    //     HBox
    //       "OK"
    //       UISpacer
    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Cost for building that starship"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    GCost have;
    have.set(el_Money, planet.getCargo(el_Money));
    have.set(el_Supplies, planet.getCargo(el_Supplies));
    have.set(el_Tritanium, planet.getCargo(el_Tritanium));
    have.set(el_Duranium, planet.getCargo(el_Duranium));
    have.set(el_Molybdenum, planet.getCargo(el_Molybdenum));

    UIGroup& g0 = h.add(new UIGroup(UIVBoxLayout::instance0));
    g0.add(h.add(new WBillDisplay(1000, bill.size(), false, bill)));
    g0.add(h.add(new WBillTotalDisplay(have, bill)));
    win.add(g0);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g1.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g1.add(h.add(new UISpacer()));
    win.add(g1);
    win.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));
    win.pack();

    ui_root->centerWidget(win);
    ui_root->runModal(win);
}
