/**
  *  \file client/dlg-fcode.cc
  *
  *  Friendly-code input dialog box.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-fcode.h"
#include "client/widgets/helpwidget.h"
#include "game/racename.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"

WFCodeListbox::WFCodeListbox(int id, int lines, int width, const GFCodeList& list)
    : UIListbox(id, lines, width, list.getSize()), list(list)
{ }

WFCodeListbox::~WFCodeListbox()
{ }

void
WFCodeListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    const int fcwidth = 3 * font_ems[FONT_NORMAL];
    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, line, &area);
        if (index < list.getSize()) {
            area.consumeX(5);
            outTextF(ctx, area.splitX(fcwidth), list[index].getFCode());
            outTextF(ctx, area, player_racenames.expandNames(list[index].getDescription()));
        }
    }
}

void
WFCodeListbox::setFCode(const string_t& fc)
{
    GFCodeList::iterator i = list.getFCodeByName(fc);
    if (i != list.end())
        scrollTo(i - list.begin(), 1);
}

string_t
WFCodeListbox::getFCode() const
{
    if (getCurrent() < list.getSize())
        return list[getCurrent()].getFCode();
    else
        return string_t();
}

/****************************** WFCodeWindow *****************************/

WFCodeWindow::WFCodeWindow(const string_t& title, const GFCodeList& list)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      input(h.add(new UIInputLine(3, 10, ss_Alt + 'f', UIInputLine::ilf_NoHi, 1000))),
      fclist(h.add(new WFCodeListbox(1001, 10, 380, list)))
{
    init();
}

void
WFCodeWindow::init()
{
    // VBox
    //   HBox
    //     UIStaticText "FCode:"
    //     UIInputLine
    //   HBox
    //     WFCodeListbox
    //     UIScrollbar
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UIButton "Alt-R Random"
    //     UISpacer
    //     UIButton "Help"

    setLayout(UIVBoxLayout::instance5);
    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIFrameGroup& g2 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIStaticText(_("FCode:"), UIStaticText::StaticTitle)));
    g1.add(input);

    g2.add(fclist);
    g2.add(h.add(new UIScrollbar(fclist)));

    g3.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    g3.add(h.add(new UIButton(0, _("Alt-R - Random"), ss_Alt + 'r')).addCall(this, &WFCodeWindow::onRandom));
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(_("Help"), ss_Alt + 'h')));

    add(g1);
    add(g2);
    add(g3);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));
    add(h.add(new WHelpWidget("pcc2:fcode")));
    pack();

    fclist.sig_moved.add(this, &WFCodeWindow::scrolled);
    input.sig_changed.add(this, &WFCodeWindow::edited);
}

WFCodeWindow::~WFCodeWindow()
{ }

void
WFCodeWindow::onRandom(int)
{
    input.setText(GFCode::generateRandomFCode());
    input.setFocusedWidget();
}

void
WFCodeWindow::edited(int)
{
    fclist.setFCode(input.getText());
}

void
WFCodeWindow::scrolled(int)
{
    if (fclist.hasState(st_Focused))
        input.setText(fclist.getFCode());
}

void
WFCodeWindow::setFCode(const string_t& fc)
{
    input.setText(fc);         // triggers /edited/
}
