/**
  *  \file flak/flak.cc
  *  \brief Fleet Action Kombat
  *  \ingroup FLAK
  *
  *  This contains the main definitions for Fleet Action Kombat, a
  *  multi-ship combat system jointly developed by the Planetsserver /
  *  Vagabond / PHost crew.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include <iostream>
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "flak/flak.h"
#include "flak/flak-port.h"
#include "game/formula.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"

#define FLAK_LOGGING

/** \defgroup FLAK Fleet Aktion Combat
    FLAK is a multi-ship combat system jointly developed by the Planetsserver /
    Vagabond / PHost crew. */

/** Magic number for FLAK VCR files. */
const char FLAK_MAGIC[8] = { 'F','L','A','K','V','C','R',26 };

const FlakBattle::ship_t FlakBattle::no_ship;
const FlakBattle::fleet_t FlakBattle::no_fleet;

/******************************** Logging ********************************/

#ifdef FLAK_LOGGING
#include "cpluslib/format.h"
static std::map<const char*, int> log_counter;

static void
clearFlakLog()
{
    std::map<const char*,int>().swap(log_counter);
}

static void
addFlakLog(const char* name, int value = 1)
{
    log_counter[name] += value;
}

string_t
getFlakLog()
{
    string_t accum;
    for (std::map<const char*,int>::iterator i = log_counter.begin(), end = log_counter.end(); i != end; ++i)
    {
        accum += format("%s: %d\n", i->first, i->second);
    }
    return accum;
}
#else
static inline void clearFlakLog() { }
static inline void addFlakLog(const char*, int = 1) { }

string_t
getFlakLog()
{
    return string_t();
}
#endif

/**************************** Simple Formulas ****************************/

// NOTE: some of this is duplicated from game/pvcr.cc.

static int
limit(int value, int min, int max)
{
    return value < min ? min : value > max ? max : value;
}

/// Compute bay recharge rate.
static inline int
computeBayRechargeRate(int bay_count, int level, int pid)
{
    int val = CONFIG_EX(BayRechargeRate, level, pid, 0, 16384) +
        bay_count * CONFIG_EX(BayRechargeBonus, level, pid, -500, 500);
    return val < 1 ? 1 : val;
}

/// Compute beam recharge rate.
static inline int
computeBeamRechargeRate(int beam_type, int level, int pid)
{
    int smash = getBeam(beam_type).getKill() + getBeam(beam_type).getDamage();
    int val = CONFIG_EX(BeamRechargeRate, level, pid, 0, 16384)
        + smash * CONFIG_EX(BeamRechargeBonus, level, pid, -4095, 4095) / 100;
    return val < 1 ? 1 : val;
}

/// Compute beam hit odds.
static inline int
computeBeamHitOdds(int beam_type, int level, int pid)
{
    int smash = getBeam(beam_type).getKill() + getBeam(beam_type).getDamage();
    int val = CONFIG_EX(BeamHitOdds, level, pid, 0, 100)
        + smash * CONFIG_EX(BeamHitBonus, level, pid, -4095, 4095) / 100;
    return val < 0 ? 0 : val;
}

/// Compute torpedo tube recharge rate.
static inline int
computeTubeRechargeRate(int torp_type, int level, int pid)
{
    int smash = getTorp(torp_type).getKill() + getTorp(torp_type).getDamage();
    int val = CONFIG_EX(TubeRechargeRate, level, pid, 0, 16384)
        + smash * CONFIG_EX(TubeRechargeBonus, level, pid, -4095, 4095) / 100;
    return val < 1 ? 1 : val;
}

/// Compute torp hit odds.
static inline int
computeTorpHitOdds(int torp_type, int level, int pid)
{
    int smash = getTorp(torp_type).getKill() + getTorp(torp_type).getDamage();
    int val = CONFIG_EX(TorpHitOdds, level, pid, 0, 100)
        + smash * CONFIG_EX(TorpHitBonus, level, pid, -4095, 4095) / 100;
    return val < 0 ? 0 : val;
}

/** Compute shield damage.
    \param expl,kill  weapon parameters
    \param mass       mass of unit being hit
    \param who        unit which fires the weapon */
static double
computeShieldDamage(int expl, int kill, int mass, const FlakShip& who)
{
    double damage = (who.config.ShieldKillScaling * double(kill)
                     + who.config.ShieldDamageScaling * double(expl)) / (mass + 1);
    return damage > 10000
        ? 10000
        : CONFIG_SCALAR(AllowAlternativeCombat)
        ? damage
        : int(damage + 1.5);
}

/** Compute hull damage.
    \param expl,kill  weapon parameters
    \param mass       mass of unit being hit
    \param who        unit which fires the weapon */
static double
computeHullDamage(int expl, int kill, int mass, const FlakShip& who)
{
    if (CONFIG_SCALAR(AllowAlternativeCombat)) {
        double d = double(expl) * who.config.HullDamageScaling / (mass+1);
        return d > 10000 ? 10000 : d;
    } else {
        double d = computeShieldDamage(expl, kill, mass, who) * who.config.HullDamageScaling / (mass+1);
        return d > 10000 ? 10000 : int(d + 1.5);
    }
}

/** Compute crew killed.
    \param expl,kill  weapon parameters
    \param death_flag zero if this is a death ray, nonzero otherwise
    \param mass       mass of unit being hit
    \param who        unit which fires the weapon */
static double
computeCrewKilled(int kill, int mass, int death_flag, const FlakShip& who)
{
    double d = double(kill) * who.config.CrewKillScaling / (mass+1);
    if (CONFIG_SCALAR(AllowAlternativeCombat)) {
        return d;
    } else {
        long el = long(d + 0.5);
        if (el == 0 && death_flag == 0)
            return 1.0;
        else
            return el;
    }
}

/******************************** FlakPos ********************************/

/** Compute distance between two positions. FLAK is essentially a 2-D
    system, so we ignore the Z coordinates completely. */
double
FlakPos::distanceTo(const FlakPos& other) const
{
    /* these must be double, we're computing stuff like 200000^2 here */
    register const double dist_x = x - other.x;
    register const double dist_y = y - other.y;
    return std::sqrt(dist_x*dist_x + dist_y*dist_y);
}

/** Check whether distance between two positions is smaller than distance.
    \param a,b  positions
    \param dist distance to check, should be in 16-bit range so that it
                does not overflow */
bool
FlakPos::isDistanceLERadius(const FlakPos& other, int32 dist) const
{
    register const int32 dist_x = std::abs(x - other.x);
    register const int32 dist_y = std::abs(y - other.y);

    return dist_x <= dist && dist_y <= dist
        && dist_x*dist_x + dist_y*dist_y <= dist*dist;
}

/******************************** FlakShip *******************************/

/*! \class FlakShip
    This represents a participant of a FLAK fight. */

/** Constructor.
    \param fleet_link   the fleet I'm in
    \param battle_link  the battle I'm in
    \param data         data from file */
FlakShip::FlakShip(FlakFleet* fleet_link, FlakBattle* battle_link, const TFlakShip& data)
    : data(data), fleet_link(fleet_link), battle_link(battle_link)
{
    init(0);
}

/** Destructor. */
FlakShip::~FlakShip()
{ }

/** Initialize the status/config structures from things given in data.
    Note that this will overwrite this ship's z position.
    \param number this ship's number */
void
FlakShip::init(int number)
{
    status.shield            = data.shield_init;
    status.damage            = data.damage_init;
    status.crew              = data.crew_init;
    status.is_alive          = true;
    status.last_hit_by       = 0;
    status.torp_limit        = data.torp_lcount;
    status.launch_countdown  = 0;
    status.fighters_launched = 0;
    status.z_offs            = 0;   // will be computed elsewhere
    status.fighter_count     = data.fighter_count;
    status.torp_count        = data.torp_count;
    status.number            = number;

    config.StandoffDistance = CONFIG_SCALAR(StandoffDistance);
    config.BayRechargeRate  = computeBayRechargeRate(data.bay_count, data.level, data.player);
    if (data.beam_count) {
        config.BeamRechargeRate = computeBeamRechargeRate(data.beam_type, data.level, data.player);
        config.BeamHitOdds      = computeBeamHitOdds(data.beam_type, data.level, data.player);
    } else {
        config.BeamRechargeRate = config.BeamHitOdds = 0;
    }

    if (data.torp_lcount) {
        config.TubeRechargeRate = computeTubeRechargeRate(data.torp_type, data.level, data.player);
        config.TorpHitOdds      = computeTorpHitOdds(data.torp_type, data.level, data.player);
    } else {
        config.TubeRechargeRate = config.TorpHitOdds = 0;
    }

    config.ShieldKillScaling    = CONFIG_EX(ShieldKillScaling,       data.level, data.player, 0, 32767);
    config.ShieldDamageScaling  = CONFIG_EX(ShieldDamageScaling,     data.level, data.player, 0, 32767);
    config.CrewKillScaling      = CONFIG_EX(CrewKillScaling,         data.level, data.player, 0, 32767);
    config.HullDamageScaling    = CONFIG_EX(HullDamageScaling,       data.level, data.player, 0, 32767);
    config.StrikesPerFighter    = CONFIG_EX(StrikesPerFighter,       data.level, data.player, 1, 100);
    config.FighterMovementSpeed = CONFIG_EX(FighterMovementSpeed,    data.level, data.player, 1, 10000);
    config.FighterBeamKill      = CONFIG_EX(FighterBeamKill,         data.level, data.player, 1, 1000);
    config.FighterBeamExplosive = CONFIG_EX(FighterBeamExplosive,    data.level, data.player, 1, 1000);
    config.BayLaunchInterval    = CONFIG_ARRAY(BayLaunchInterval,    data.player);
    config.TorpFiringRange      = CONFIG_ARRAY(TorpFiringRange,      data.player);
    config.BeamHitFighterCharge = CONFIG_EX(BeamHitFighterCharge,    data.level, data.player, 0, 1000);
    config.BeamHitShipCharge    = CONFIG_ARRAY(BeamHitShipCharge,    data.player);
    config.BeamFiringRange      = CONFIG_ARRAY(BeamFiringRange,      data.player);
    config.FireOnAttackFighters = CONFIG_SCALAR(FireOnAttackFighters);
    config.FighterKillOdds      = CONFIG_ARRAY(FighterKillOdds,      data.player);
    config.FighterFiringRange   = CONFIG_ARRAY(FighterFiringRange,   data.player);

    int st = data.shield_init == 100 ? 1000 : 0;
    for (unsigned i = 0; i < countof(status.beam_charge); ++i)
        status.beam_charge[i] = st;
    for (unsigned i = 0; i < countof(status.torp_charge); ++i)
        status.torp_charge[i] = st;
    for (unsigned i = 0; i < countof(status.bay_charge); ++i)
        status.bay_charge[i] = st;
    status.torp_rx_count = 0;

    status.min_fighters_aboard = status.fighter_count;
    status.torps_hit = 0;
}

/** Weapon recharge work. */
void
FlakShip::recharge()
{
    if (!isAlive())
        return;
    if (status.launch_countdown > 0)
        --status.launch_countdown;
#if FLAK_ALGORITHM >= 20060531
    if (status.torp_rx_count > 0) {
        status.torp_rx_count--;
        status.torp_count++;
    }
#endif
    for (int i = 0; i < data.bay_count; ++i)
        if (status.bay_charge[i] < 1000)
            status.bay_charge[i] += battle_link->random(config.BayRechargeRate);
    for (int i = 0; i < data.beam_count; ++i)
        if (status.beam_charge[i] < 1000)
            status.beam_charge[i] += battle_link->random(config.BeamRechargeRate);
#if FLAK_ALGORITHM >= 20060531
    if (status.torp_count)
#endif
        for (int i = 0; i < data.torp_lcount; ++i)
            if (status.torp_charge[i] < 1000)
                status.torp_charge[i] += battle_link->random(config.TubeRechargeRate);
}

/******************************* FlakPlayer ******************************/

/*! \class FlakPlayer
    Contains information about one player's status. In particular, moving objects
    are managed here. */

/** Construct empty player. Since we store these objects in an array,
    we need a default constructor. */
FlakPlayer::FlakPlayer()
    : number(0), num_live_ships(0), sum_strength(0), stuff(0), have_any_fighters(false)
{ }

/** Copy constructor. This copies all information from the other
    object, and duplicates the moving objects. */
FlakPlayer::FlakPlayer(const FlakPlayer& other)
    : number(other.number), num_live_ships(other.num_live_ships),
      sum_strength(other.sum_strength),
      stuff(FlakObject::cloneList(other.stuff)),
      have_any_fighters(other.have_any_fighters)
{ }

/** Copy operator. You can only copy matching FlakPlayer objects, i.e.
    those which have a matching player number, to avoid major catastrophes. */
FlakPlayer&
FlakPlayer::operator=(const FlakPlayer& other)
{
    ASSERT(number == other.number);
    num_live_ships    = other.num_live_ships;
    sum_strength      = other.sum_strength;
    have_any_fighters = other.have_any_fighters;
    FlakObject::freeList(stuff);
    stuff = FlakObject::cloneList(other.stuff);
    return *this;
}

/** Destructor. Discards all objects. */
FlakPlayer::~FlakPlayer()
{
    FlakObject::freeList(stuff);
}

/** Add an object. */
inline void
FlakPlayer::addObject(FlakObject* p)
{
    p->next = stuff;
    stuff   = p;
}

/** Garbage collection. Deletes all objects marked for deletion. */
void
FlakPlayer::doPlayerGC()
{
    FlakObject** pp = &stuff;
    while (FlakObject* p = *pp) {
        if (p->kind == FlakObject::oDeleteMe) {
            *pp = p->next;
            delete p;
        } else {
            pp = &p->next;
        }
    }
}

/** Bring this player into virgin state. Deletes all objects and
    clears counters.
    \param number this player's number to be stored in the number field */
inline void
FlakPlayer::init(int number)
{
    FlakObject::freeList(stuff);
    num_live_ships = 0;
    sum_strength = 0;
    stuff = 0;
    have_any_fighters = false;
    this->number = number;
}

/******************************* FlakObject ******************************/

/** Move object towards to_pos. This uses the configured speed of the object.
    \returns remaining distance (zero if to_pos reached). */
int32
FlakObject::moveTowards(FlakPos to_pos)
{
    double dist = position.distanceTo(to_pos);
    double new_dist = dist - speed;
    if (new_dist <= 0) {
        position = to_pos;
        return 0;
    } else {
        position.x = to_pos.x + roundToInt(double(position.x - to_pos.x) * new_dist / dist);
        position.y = to_pos.y + roundToInt(double(position.y - to_pos.y) * new_dist / dist);
        position.z = to_pos.z + roundToInt(double(position.z - to_pos.z) * new_dist / dist);
        return int32(new_dist);
    }
}

/** Clone the given object list.
    \return new list */
FlakObject*
FlakObject::cloneList(FlakObject* p)
{
    // FIXME: exception safety
    FlakObject** q = &p;
    while (*q) {
        *q = new FlakObject(**q);
        q = &(*q)->next;
    }
    return p;
}

/** Free object list. */
void
FlakObject::freeList(FlakObject* p)
{
    while (FlakObject* q = p) {
        p = p->next;
        delete q;
    }
}

/******************************* FlakFleet *******************************/

FlakFleet::FlakFleet(const TFlakFleet& data)
    : data(data)
{
    init();
}

void
FlakFleet::init()
{
    status.enemy_ptr = 0;
    status.alive = true;
    status.position.x = data.x_init;
    status.position.y = data.y_init;
    status.position.z = 0;
}

/**************************** FlakStatusToken ****************************/

/** Initialize from battle. This copies the status of the specified
    battle. */
FlakStatusToken::FlakStatusToken(const FlakBattle& battle)
    : fleets(battle.fleets.size()), ships(battle.ships.size()),
      seed(battle.seed), time(battle.time), is_terminated(battle.is_terminated)
{
    for (std::size_t i = 0; i < battle.fleets.size(); ++i)
        fleets[i] = new FlakFleet::Status(battle.fleets[i]->status);
    for (std::size_t i = 0; i < battle.ships.size(); ++i)
        ships[i] = new FlakShip::Status(battle.ships[i]->status);
    for (std::size_t i = 0; i < countof(player_status); ++i) {
        player_status[i].init(i+1);
        player_status[i] = battle.player_status[i];
    }
}

/** Copy a status token. */
FlakStatusToken::FlakStatusToken(const FlakStatusToken& token)
    : fleets(token.fleets.size()), ships(token.ships.size()),
      seed(token.seed), time(token.time), is_terminated(token.is_terminated)
{
    for (std::size_t i = 0; i < token.fleets.size(); ++i)
        fleets[i] = new FlakFleet::Status(*token.fleets[i]);
    for (std::size_t i = 0; i < token.ships.size(); ++i)
        ships[i] = new FlakShip::Status(*token.ships[i]);
    for (std::size_t i = 0; i < countof(player_status); ++i)
        player_status[i] = token.player_status[i];
}

/** Destructor. */
FlakStatusToken::~FlakStatusToken()
{
    safeClear(fleets);
    safeClear(ships);
}

/** Store status token back to battle. After this call, the battle
    continues at the place where this token was created. This does not
    update the battle's visualisation. */
void
FlakStatusToken::storeTo(FlakBattle& battle) const
{
    ASSERT(battle.ships.size() == ships.size());
    ASSERT(battle.fleets.size() == fleets.size());

    for (std::size_t i = 0; i < battle.fleets.size(); ++i)
        battle.fleets[i]->status = *fleets[i];
    for (std::size_t i = 0; i < battle.ships.size(); ++i)
        battle.ships[i]->status = *ships[i];
    for (std::size_t i = 0; i < countof(player_status); ++i)
        battle.player_status[i] = player_status[i];
    battle.seed = seed;
    battle.time = time;
    battle.is_terminated = is_terminated;
}

/******************************* FlakBattle ******************************/

/** Compute torpedo launch limit for attacker.
    \param attacker the ship whose limit we're thinking about. The limit will be
                    updated in-place
    \param enemy    desired enemy
    \param num_torpers total number of active torpers in attacking fleet */
void
FlakBattle::computeTorpLimit(FlakShip& attacker, const FlakShip& enemy, int num_torpers)
{
    attacker.status.torp_limit = attacker.data.torp_lcount;

    /* don't bother for small ships or pathological case */
    if (attacker.data.torp_lcount <= 2 || attacker.config.TorpHitOdds <= 0)
        return;

    int expl = getTorp(attacker.data.torp_type).getDamage();
    int kill = getTorp(attacker.data.torp_type).getKill();
    if (!CONFIG_SCALAR(AllowAlternativeCombat))
        expl *= 2, kill *= 2;

    double cd = computeCrewKilled(kill, enemy.data.mass, expl /* death flag */, enemy);
    int torps_reqd;
    if (expl == 0) {
        /* death ray */
        if (enemy.isPlanet())             // should not happen
            return;
        // 'crew/cd' is number of hits required to kill crew
        // 'crew/cd * 100/TorpHitOdds' is average number of torps to kill crew
        // add safety factor
        torps_reqd = int(1 + (enemy.status.crew / cd) * FLAK_TORP_LIMIT_FACTOR / attacker.config.TorpHitOdds);
    } else {
        /* normal weapon */
        register double hd = computeHullDamage(expl, kill, enemy.data.mass, enemy);
        register double sd = computeShieldDamage(expl, kill, enemy.data.mass, enemy);

        register int limit = (raceId(enemy.data.player) == 2) ? 151 : 100;
        register double v1 = (limit - enemy.status.damage) / (hd + 0.01);
        if (!enemy.isPlanet()) {
            register double v2 = enemy.status.crew / (cd + 0.01);
            if (v1 > v2)
                v1 = v2;
        }
        torps_reqd = int(1 + (enemy.status.shield / sd + v1) * FLAK_TORP_LIMIT_FACTOR / attacker.config.TorpHitOdds);
    }

    if (num_torpers)
        torps_reqd = (torps_reqd + num_torpers-1) / num_torpers;

    if (torps_reqd < attacker.data.torp_lcount)
        attacker.status.torp_limit = torps_reqd;
}

/** Inflict damage to a unit.
    \param sh           target unit
    \param firing_ship  the ship which is responsible for hitting this ship
    \param expl,kill    weapon parameters
    \param death_flag   zero if it is a death ray */
void
FlakBattle::hitShipWith(FlakShip& sh, const FlakShip& firing_ship, int expl,
                        int kill, int death_flag)
{
    const int att_count = player_status[firing_ship.data.player-1].num_live_ships;
    const int opp_count = player_status[sh.data.player-1].num_live_ships;
    const int32 att_strength = player_status[firing_ship.data.player-1].sum_strength;
    const int32 opp_strength = player_status[sh.data.player-1].sum_strength;

    /* compute effective mass */
    int eff_mass;
    if (att_count < opp_count && att_strength < opp_strength) {
        if ((att_strength + FLAK_COMPENSATION_DIVISOR) * FLAK_COMPENSATION_LIMIT < (opp_strength + FLAK_COMPENSATION_DIVISOR))
            eff_mass = sh.data.mass / FLAK_COMPENSATION_LIMIT;
        else
            eff_mass = sh.data.mass * (att_strength+FLAK_COMPENSATION_DIVISOR) / (opp_strength+FLAK_COMPENSATION_DIVISOR);
    } else {
        eff_mass = sh.data.mass;
    }

    /* normal PHost damage processing */
    double damage_rate = 1.0;
    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;
    if (death_flag == 0)
        goto death_ray;

    if (sh.status.shield > 0) {
        double damage = computeShieldDamage(expl, kill, eff_mass, sh);
        if (sh.status.shield <= damage) {
            /* shield completely down */
            damage_rate = (damage - sh.status.shield) / damage;
            sh.status.shield = 0;
        } else {
            /* Shields still hold */
            damage_rate = 0.0;
            sh.status.shield -= damage;
        }
    }

    if (damage_rate <= 0)
        goto vis;

    sh.status.damage += computeHullDamage(expl, kill, eff_mass, sh) * damage_rate;
    if (sh.status.damage > 9999)
        sh.status.damage = 9999;

 death_ray:
    if (!sh.isPlanet()) {
        sh.status.crew -= computeCrewKilled(kill, eff_mass, death_flag, sh) * damage_rate;
        if (sh.status.crew < 0.5)
            sh.status.crew = 0;
    }

 vis:
    vis->hitShip(sh.status.number, expl, kill, death_flag);
    if (sh.isAlive())
        sh.status.last_hit_by = &firing_ship;
}

/** Check whether /we/ can attack /they/. */
static bool
canStillAttack(FlakShip& we, FlakShip& they)
{
    int torpc = we.data.torp_lcount;
    int beamc = we.data.beam_count;

    if (they.isPlanet()) {
        /* discount death rays against planet */
        if (torpc && getTorp(we.data.torp_type).getDamage() == 0)
            torpc = 0;
        if (beamc && getBeam(we.data.beam_type).getDamage() == 0
            && they.status.fighter_count == 0
            && they.status.fighters_launched == 0)
            beamc = 0;
    }

    return beamc != 0
        || (torpc != 0 && we.status.torp_count != 0)
        || (we.data.bay_count != 0
            && (we.status.fighter_count != 0
                || we.status.fighters_launched != 0));
}

/** Pick a new enemy for /fleet/. Updates the fleet in-place. */
void
FlakBattle::chooseEnemy(FlakFleet& fleet)
{
    if (!fleet.isAlive())
        return;

    int32      best_diff = 0x7FFFFFFF;
    FlakShip*  best_choice = 0;
    for (int16 index = 0; index < fleet.data.att_list_size; ++index) {
        /* bail out early if we know we cannot attack it */
        const int their_ship_nr = attack_list[2*fleet.data.att_list_pointer + 2*index];
        FlakShip& their_ship = *ships[their_ship_nr];
        if (their_ship.data.player == fleet.data.player)
            continue;
        if (!their_ship.isAlive())
            continue;

        /* figure out attack ratings */
        int32 attack_rating    = 0;
        int32 eff_rating_bonus = attack_list[2*fleet.data.att_list_pointer + 2*index + 1];

        for (int n = 0; n < fleet.data.num_ships; ++n) {
            FlakShip& we = *ships[n + fleet.data.first_ship];
            if (we.isAlive() && (canStillAttack(we, their_ship) || canStillAttack(their_ship, we)))
                attack_rating += we.data.rating;
        }

        /* no attack rating means we cannot attack it at all */
        if (attack_rating == 0)
            continue;

        int32 their_rating = their_ship.data.rating;
        int32 divisor, diff;

        if (attack_rating < their_rating) {
            /* we're smaller */
            diff = their_rating - attack_rating + FLAK_DIFF_OFFSET;
            divisor = FLAK_DIVISOR_IF_SMALLER;
        } else {
            /* we're bigger */
            diff = attack_rating - their_rating + FLAK_DIFF_OFFSET;
            divisor = FLAK_DIVISOR_IF_BIGGER;
        }

        /* bonuses */
        divisor += eff_rating_bonus;
        if (&their_ship == fleet.status.enemy_ptr)
            divisor += FLAK_DIVISOR_SAME_ENEMY_BONUS;
        if (their_ship.status.damage > 0)
            divisor += int(their_ship.status.damage);
        divisor += (100 - int(their_ship.status.shield)) / 5;

        int32 multiplier = int32(fleet.data.speed <= 0
                                 ? fleet.status.position.distanceTo(their_ship.fleet_link->status.position) / 100
                                 : fleet.status.position.distanceTo(their_ship.fleet_link->status.position) / fleet.data.speed);
        if (multiplier < FLAK_MULTIPLIER_MIN)
            multiplier = FLAK_MULTIPLIER_MIN;

        diff = diff * multiplier / divisor;
        if (diff < best_diff) {
            best_diff = diff;
            best_choice = &their_ship;
        }
    }

    if (best_choice != fleet.status.enemy_ptr && best_choice != 0 && fleet.status.enemy_ptr != 0)
        addFlakLog("Target change in flight");

    fleet.status.enemy_ptr = best_choice;

    if (best_choice) {
        /* re-compute torp limit */
        int num_torpers = 0;
        for (int i = 0; i < fleet.data.num_ships; ++i) {
            FlakShip& sh = *ships[i+fleet.data.first_ship];
            if (sh.isAlive() && sh.data.torp_lcount && sh.status.torp_count >= 10)
                ++num_torpers;
        }
        for (int i = 0; i < fleet.data.num_ships; ++i)
            computeTorpLimit(*ships[i+fleet.data.first_ship], *best_choice, num_torpers);
    }
}

/** Launch fighters for given fleet. */
void
FlakBattle::launchFighters(const FlakFleet& fleet)
{
    if (!fleet.isAlive() || !fleet.status.enemy_ptr || !fleet.status.enemy_ptr->isAlive())
        return;

    const ship_t last = fleet.data.first_ship + fleet.data.num_ships;
    for (ship_t i = fleet.data.first_ship; i < last; ++i) {
        FlakShip& sh = *ships[i];
        if (sh.isAlive() && sh.data.bay_count && sh.status.fighter_count
            && sh.status.launch_countdown==0 && sh.status.fighters_launched < sh.data.max_fighters_launched)
        {
            /* we can launch a fighter. Find a bay. */
            for (int bay = 0; bay < sh.data.bay_count; ++bay) {
                if (sh.status.bay_charge[bay] < 1000)
                    continue;

                FlakObject& p = *new FlakObject();

                p.kind             = FlakObject::oFighter;
                p.can_change_enemy = true;
                p.position         = sh.getPos();
                p.enemy_ptr        = fleet.status.enemy_ptr;
                p.owner_ptr        = &sh;
                p.strikes          = sh.config.StrikesPerFighter;
                p.kill             = sh.config.FighterBeamKill;
                p.expl             = sh.config.FighterBeamExplosive;
                p.death_flag       = 1;
                p.speed            = sh.config.FighterMovementSpeed;

                sh.status.bay_charge[bay] = 0;
                sh.status.fighter_count--;
                sh.status.fighters_launched++;
                sh.status.launch_countdown = sh.config.BayLaunchInterval;
                if (sh.status.fighter_count < sh.status.min_fighters_aboard)
                    sh.status.min_fighters_aboard = sh.status.fighter_count;

                player_status[sh.data.player-1].addObject(&p);
                player_status[sh.data.player-1].have_any_fighters = true;
                vis->createObject(p, i);
                goto next_ship;
            }
        }
     next_ship:;
    }
}

/** Fire torps from /fleet/. */
void
FlakBattle::fireTorps(const FlakFleet& fleet)
{
    if (!fleet.isAlive() || !fleet.status.enemy_ptr || !fleet.status.enemy_ptr->isAlive())
        return;

    const FlakPos& ene_pos = fleet.status.enemy_ptr->fleet_link->status.position;
    const FlakPos& my_pos  = fleet.status.position;

    const ship_t last = fleet.data.num_ships + fleet.data.first_ship;
    for (ship_t i = fleet.data.first_ship; i < last; ++i) {
        FlakShip& sh = *ships[i];
        if (!sh.isAlive() || !sh.status.torp_count)
            continue;           // torp cannot fire
        if (!my_pos.isDistanceLERadius(ene_pos, sh.config.TorpFiringRange))
            continue;

        for (int tl = 0; tl < sh.status.torp_limit; ++tl) {
            if (sh.status.torp_charge[tl] < 1000)
                continue;

            FlakObject& p = *new FlakObject();

            p.kind             = FlakObject::oTorpedo;
            p.can_change_enemy = false;
            p.position         = sh.getPos();
            p.enemy_ptr        = fleet.status.enemy_ptr;
            p.owner_ptr        = &sh;
            p.strikes          = 0;
            p.kill             = getTorp(sh.data.torp_type).getKill();
            p.expl             = getTorp(sh.data.torp_type).getDamage();
            p.death_flag       = p.expl;
            p.speed            = FLAK_TORP_MOVEMENT_SPEED;
            if (!CONFIG_SCALAR(AllowAlternativeCombat))
                p.kill *= 2, p.expl *= 2;
            if (random(100) < sh.config.TorpHitOdds)
                p.strikes = 1;

            sh.status.torp_charge[tl] = 0;
            sh.status.torp_count--;

            player_status[sh.data.player-1].addObject(&p);
            vis->createObject(p, i);
            goto next_ship;
        }
     next_ship:;
    }
}

/** Fire all beams from fleet. */
void
FlakBattle::fireBeams(const FlakFleet& fleet)
{
    if (!fleet.isAlive())
        return;

    const double dist = fleet.status.enemy_ptr
        ? fleet.status.position.distanceTo(fleet.status.enemy_ptr->fleet_link->status.position)
        : 1.0e+15;

    const int last_ship = fleet.data.first_ship + fleet.data.num_ships;
    for (int sh = fleet.data.first_ship; sh < last_ship; ++sh) {
        FlakShip& ship = *ships[sh];
        if (!ship.isAlive())
            continue;

        /* first, try to fire at fighters */
        for (int bm = 0; bm < ship.data.beam_count; ++bm) {
            if (ship.status.beam_charge[bm] >= ship.config.BeamHitFighterCharge) {
                /* find a fighter */
                FlakObject* min_ftr = 0;
                double min_dist;
                for (int pl = 1; pl <= int(countof(player_status)); ++pl) {
                    if (pl == ship.data.player)
                        continue;
                    for (FlakObject* p = player_status[pl-1].stuff; p != 0; p = p->next) {
                        if (p->kind == FlakObject::oFighter
                            && p->enemy_ptr && p->owner_ptr && fleet.status.enemy_ptr
                            && (p->enemy_ptr->fleet_link == ship.fleet_link
                                || p->owner_ptr->fleet_link == fleet.status.enemy_ptr->fleet_link))
                        {
                            double fdist = fleet.status.position.distanceTo(p->position);
                            if (fdist <= ship.config.BeamFiringRange
                                && (!min_ftr
                                    || (CONFIG_SCALAR(FireOnAttackFighters)
                                        && min_ftr->strikes == 0 && p->strikes > 0)
#if FLAK_ALGORITHM >= 20060531
                                    || (fdist <= min_dist
#else
                                    || (fdist < min_dist
#endif
                                        && (!CONFIG_SCALAR(FireOnAttackFighters)
                                            || (min_ftr->strikes == 0 && p->strikes == 0)
                                            || (min_ftr->strikes > 0 && p->strikes > 0)))))
                            {
                                min_ftr = p;
                                min_dist = fdist;
                            }
                        }
                    }
                }

                /* got a fighter? */
                if (min_ftr) {
                    if (random(100) < ship.config.BeamHitOdds) {
                        vis->fireBeam(sh, *min_ftr, true);
                        vis->destroyObject(*min_ftr, true);
                        min_ftr->kind = FlakObject::oDeleteMe;
                        if (min_ftr->owner_ptr)
                            --min_ftr->owner_ptr->status.fighters_launched;
                    } else {
                        vis->fireBeam(sh, *min_ftr, false);
                    }
                    ship.status.beam_charge[bm] = 0;
                    goto next_ship;
                }
            } /* if beam charged */
        } /* for each beam */

        if (fleet.status.enemy_ptr && fleet.status.enemy_ptr->isAlive()
            && dist <= ship.config.BeamFiringRange)
        {
            for (int bm = 0; bm < ship.data.beam_count; ++bm) {
                if (ship.status.beam_charge[bm] >= ship.config.BeamHitShipCharge) {
                    int kill = getBeam(ship.data.beam_type).getKill();
                    int damage = getBeam(ship.data.beam_type).getDamage();
                    if (raceId(ship.data.player) == 5)
                        kill *= 3;
                    if (random(100) < ship.config.BeamHitOdds) {
                        vis->fireBeam(sh, fleet.status.enemy_ptr->status.number, true);
                        hitShipWith(*fleet.status.enemy_ptr, ship,
                                    damage * long(ship.status.beam_charge[bm]) / 1000,
                                    kill * long(ship.status.beam_charge[bm]) / 1000,
                                    damage /* death flag */);
                    } else {
                        vis->fireBeam(sh, fleet.status.enemy_ptr->status.number, false);
                    }
                    ship.status.beam_charge[bm] = 0;
                    goto next_ship;
                }
            }
        }

     next_ship:;
    }
}

/** Check whether battle has ended.
    \return true if battle ends. */
bool
FlakBattle::endCheck() const
{
    for (unsigned i = 0; i < countof(player_status); ++i)
        if (player_status[i].stuff)
            return false;
    for (unsigned i = 0; i < fleets.size(); ++i)
        if (fleets[i]->status.alive && fleets[i]->status.enemy_ptr)
            return false;
    return true;
}

/** Compute new position for a fleet. Modifies the fleet in-place.
    This computes the new_position field, it does not actually move the
    fleet. */
void
FlakBattle::computeNewPosition(FlakFleet& fleet)
{
    fleet.new_position = fleet.status.position;

    if (!fleet.isAlive())
        return;

    /* if our enemy died, pick a new one */
    if (fleet.status.enemy_ptr && !fleet.status.enemy_ptr->isAlive())
        chooseEnemy(fleet);

    /* FIXME: this is not quite clean because it gets the StandoffDistance
       from the ship. Better idea? */
    const int sod = ships[fleet.data.first_ship]->config.StandoffDistance;
    if (fleet.status.enemy_ptr) {
        /* move towards target */
        FlakPos target = fleet.status.enemy_ptr->fleet_link->status.position;

        double dist = fleet.status.position.distanceTo(target);
        if (dist > 0) {
            double nd = dist - fleet.data.speed;
            if (nd < sod)
                nd = sod;

            if (nd < dist) {
                fleet.new_position.x = target.x + roundToInt((fleet.status.position.x - target.x) * nd / dist);
                fleet.new_position.y = target.y + roundToInt((fleet.status.position.y - target.y) * nd / dist);
                fleet.new_position.z = 0;
            }
        }
    }

    /* 'bouncing'. I know this is physical bullshit, but it looks nice. */
    const double speed = 0.75;
    int32 bvx = 0, bvy = 0;
    double bvdivi = 0;

    int player = fleet.data.player;

    for (unsigned ff = 0; ff < fleets.size(); ++ff) {
        if (fleets[ff] == &fleet || !fleets[ff]->isAlive())
            continue;

        double idist = fleets[ff]->status.position.distanceTo(fleet.new_position);
        int    their_player = fleets[ff]->data.player;
        if (idist > 10 && ((player == their_player ? 4*idist : 2*idist) < sod)) {
            bvx += fleet.new_position.x - fleets[ff]->status.position.x;
            bvy += fleet.new_position.y - fleets[ff]->status.position.y;
            bvdivi += idist;
        }
    }

    if (bvx || bvy) {
        fleet.new_position.x = roundToInt(fleet.new_position.x + bvx*speed*fleet.data.speed / bvdivi);
        fleet.new_position.y = roundToInt(fleet.new_position.y + bvy*speed*fleet.data.speed / bvdivi);
    }
}

/** Fleet GC. Deletes all ships killed this tick, and marks the fleet
    dead if it happens. */
void
FlakBattle::doFleetGC(FlakFleet& fleet)
{
    if (!fleet.isAlive())
        return;

    /* kill all dead ships */
    const int limit = raceId(fleet.data.player) == 2 ? 150 : 99;
    bool alive = false;
    bool any_torps = false;
    int torps[NUM_TORPS];
    for (int i = 0; i < fleet.data.num_ships; ++i) {
        FlakShip& sh = *ships[i + fleet.data.first_ship];
        if (sh.isAlive()) {
            if (sh.status.damage > limit || (!sh.isPlanet() && sh.status.crew < 0.5)) {
                vis->destroyShip(i + fleet.data.first_ship);
                sh.status.is_alive = false;
                --player_status[sh.data.player-1].num_live_ships;
                player_status[sh.data.player-1].sum_strength -= sh.data.compensation_rating;
#if FLAK_ALGORITHM >= 20060531
                if (sh.data.torp_type > 0 && sh.data.torp_type <= NUM_TORPS && sh.status.torp_count > 0) {
                    if (!any_torps)
                        for (int x = 0; x < NUM_TORPS; ++x)
                            torps[x] = 0;
                    any_torps = true;
                    torps[sh.data.torp_type-1] += sh.status.torp_count;
                }
#endif
            } else {
                alive = true;
            }
        }
    }
    fleet.status.alive = alive;
#if FLAK_ALGORITHM >= 20060531
    if (alive && any_torps) {
        for (int i = 0; i < fleet.data.num_ships; ++i) {
            FlakShip& sh = *ships[i + fleet.data.first_ship];
            if (sh.isAlive() && sh.data.torp_type > 0 && sh.data.torp_type <= NUM_TORPS)
                sh.status.torp_rx_count += torps[sh.data.torp_type-1] / fleet.data.num_ships;
        }
    }
#endif
}

/** Do fighter-intercept phase for two players. */
void
FlakBattle::fighterIntercept(FlakPlayer& a, FlakPlayer& b)
{
    if (CONFIG_ARRAY(FighterKillOdds, a.number) == 0 && CONFIG_ARRAY(FighterKillOdds, b.number) == 0)
        return;

    for (FlakObject* pa = a.stuff; pa != 0; pa = pa->next) {
        if (pa->kind == FlakObject::oFighter) {
            for (FlakObject* pb = b.stuff; pb != 0; pb = pb->next) {
                if (pb->kind == FlakObject::oFighter) {
                    /* two fighters. Possible targets? */
                    if (pa->owner_ptr != 0 && pb->owner_ptr != 0 &&
                        (pa->enemy_ptr == pb->owner_ptr
                         || pb->enemy_ptr == pa->owner_ptr))
                    {
                        if (tryIntercept(*pa, *pb))
                            return;
                        if (pa->kind != FlakObject::oFighter)
                            break;  /* can this happen? I think no. */
                    }
                }
            }
        }
    }
}

/** Attempt fighter-intercept between two fighters.
    \param pa,pb  two fighters
    \return true if successful, false if no fighter intercept here */
bool
FlakBattle::tryIntercept(FlakObject& pa, FlakObject& pb)
{
    /* FIXME: I'm not sure that the probabilities are correct. This
       yields one_f==0 if both are 100. */
    if (!pa.position.isDistanceLERadius(pb.position, FLAK_FIGHTER_INTERCEPT_RANGE))
        return false;

    int left_f  = (100 - pa.owner_ptr->config.FighterKillOdds) * pb.owner_ptr->config.FighterKillOdds;
    int right_f = (100 - pb.owner_ptr->config.FighterKillOdds) * pa.owner_ptr->config.FighterKillOdds;
    int one_f   = (left_f + right_f) / 100;

    if (one_f == 0)
        return false;

    int right_probab = right_f / one_f;

    if (random(100) >= one_f)
        return false;

    if (random(100) < right_probab) {
        /* a killed */
        vis->fireBeam(pb, pa);
        vis->destroyObject(pa, true);
        pa.kind = FlakObject::oDeleteMe;
        if (pa.owner_ptr)
            pa.owner_ptr->status.fighters_launched--;
        addFlakLog("Fighter Intercept A Killed");
    } else {
        /* b killed */
        vis->fireBeam(pa, pb);
        vis->destroyObject(pb, true);
        pb.kind = FlakObject::oDeleteMe;
        if (pb.owner_ptr)
            pb.owner_ptr->status.fighters_launched--;
        addFlakLog("Fighter Intercept B Killed");
    }
    return true;
}

/** Fighters of a player fire. */
void
FlakBattle::fightersFire(const FlakPlayer& player)
{
    for (FlakObject* p = player.stuff; p != 0; p = p->next) {
        if (p->kind == FlakObject::oFighter && p->strikes && p->enemy_ptr->isAlive()) {
            if (p->position.isDistanceLERadius(p->enemy_ptr->fleet_link->status.position,
                                               p->owner_ptr->config.FighterFiringRange))
            {
                vis->fireBeam(*p, p->enemy_ptr->status.number);
                hitShipWith(*p->enemy_ptr, *p->owner_ptr,
                            p->kill, p->expl, p->death_flag);
                p->strikes--;
                p->can_change_enemy = false;
            }
        }
    }
}

/** Find new base for a fighter.
    \param player      the player whose fighter we talk about
    \param fighter     the fighter which needs a new base */
void
FlakBattle::findNewBase(const FlakPlayer& player, FlakObject& fighter)
{
    double     min_dist = 1.0E+15;
    FlakShip*  min_ship = 0;

    // FIXME: somehow, we should limit this search to the player's
    // fleets to improve performance
    for (unsigned i = 0; i < fleets.size(); ++i) {
        if (fleets[i]->data.player != player.number)
            continue;
        double this_dist = fleets[i]->status.position.distanceTo(fighter.position);
        if (this_dist < min_dist) {
            int max_mass = 0;
            const ship_t last = fleets[i]->data.first_ship + fleets[i]->data.num_ships;
            for (ship_t j = fleets[i]->data.first_ship; j < last; ++j) {
                FlakShip& sh = *ships[j];
                if (sh.isAlive() && sh.data.bay_count && sh.data.mass > max_mass) {
                    max_mass = sh.data.mass;
                    min_dist = this_dist;
                    min_ship = &sh;
                }
            }
        }
    }
    if (min_ship)
        min_ship->status.fighters_launched++;
    fighter.owner_ptr = min_ship;
}

/** Move all objects belonging to a player. */
void
FlakBattle::moveStuff(FlakPlayer& player)
{
    for (FlakObject* p = player.stuff; p != 0; p = p->next) {
        if (p->kind == FlakObject::oTorpedo) {
            if (p->moveTowards(p->enemy_ptr->getPos()) == 0) {
                /* torp hits */
                if (p->strikes) {
                    FlakShip* ene = p->enemy_ptr;
                    if (!ene->isAlive() && ene->fleet_link->isAlive())
                        ene = ships[ene->fleet_link->data.first_ship + random(ene->fleet_link->data.num_ships)];
                    if (ene->isAlive()) {
                        p->enemy_ptr = ene;
                        vis->destroyObject(*p, p->strikes);
                        hitShipWith(*ene, *p->owner_ptr,
                                    p->expl, p->kill, p->death_flag);
                        p->owner_ptr->status.torps_hit++;
                        addFlakLog("Torp hitting unit");
                    } else {
                        p->enemy_ptr = 0;
                        vis->destroyObject(*p, p->strikes);
                        addFlakLog("Torp hitting void");
                    }
                } else {
                    addFlakLog("Torp missing");
                    vis->destroyObject(*p, p->strikes);
                }
                p->kind = FlakObject::oDeleteMe;
            }
        } else if (p->kind == FlakObject::oFighter) {
            if (p->strikes && !p->enemy_ptr->isAlive()) {
                /* enemy died. */
                if (!p->can_change_enemy) {
                    /* we cannot change our enemy. Retreat. */
                    addFlakLog("Fighter retreats after shooting");
                    p->strikes = 0;
                } else if (p->owner_ptr && p->owner_ptr->isAlive()) {
                    /* base is still alive. Ask it what to do. */
                    if (p->owner_ptr->fleet_link->status.enemy_ptr != p->enemy_ptr) {
                        addFlakLog("Fighter changing target");
                        p->enemy_ptr = p->owner_ptr->fleet_link->status.enemy_ptr;
#if FLAK_ALGORITHM >= 20060531
                        p->strikes >>= 1;
#endif
                    } else {
                        /* base has not decided yet. Wait. */
                        addFlakLog("Fighter waiting for base");
                        p->strikes--;
                    }
                } else {
                    /* base died. Pick new one and ask it what to do. */
                    addFlakLog("Fighter picking new base");
                    findNewBase(player, *p);
                    if (p->owner_ptr) {
                        p->enemy_ptr = p->owner_ptr->fleet_link->status.enemy_ptr;
                    } else {
                        p->enemy_ptr = 0;
                    }
                }
            }

            if (!p->enemy_ptr || player.num_live_ships == 0)
                p->strikes = 0;

            if (p->strikes == 0) {
                /* return to base */
                if (!p->owner_ptr || !p->owner_ptr->isAlive())
                    findNewBase(player, *p);

                if (!p->owner_ptr) {
                    // FIXME: I'm unsure whether to consider this 'violent' or not.
                    vis->destroyObject(*p, true);
                    p->kind = FlakObject::oDeleteMe;
                } else {
                    if (p->moveTowards(p->owner_ptr->getPos()) == 0) {
                        /* reached base */
                        vis->destroyObject(*p, false);
                        p->kind = FlakObject::oDeleteMe;
                        p->owner_ptr->status.fighter_count++;
                        p->owner_ptr->status.fighters_launched--;
                    }
                }
            } else {
                /* move towards enemy */
                p->moveTowards(p->enemy_ptr->getPos());
            }
        }
    }
}

/** Initialize everything. If battle was already advanced a bit,
    move back to beginning. */
void
FlakBattle::init()
{
    ASSERT(vis);

    time          = 0;
    seed          = orig_seed;
    is_terminated = false;

    clearFlakLog();

    for (unsigned i = 0; i < countof(player_status); ++i)
        player_status[i].init(i+1);
    for (ship_t i = 0; i < ships.size(); ++i)
        ships[i]->init(i);
    for (fleet_t i = 0; i < fleets.size(); ++i)
        fleets[i]->init();

    for (fleet_t i = 0; i < fleets.size(); ++i) {
        // FIXME: maybe merge this with FlakFleet::init() ?
        FlakFleet& fl = *fleets[i];
        int cur_z = -(fl.data.num_ships-1) * 50;
        const ship_t last = fl.data.first_ship + fl.data.num_ships;
        for (ship_t s = fl.data.first_ship; s < last; ++s) {
            ships[s]->status.z_offs = cur_z;
            cur_z += 100;

            player_status[fl.data.player-1].num_live_ships++;
            player_status[fl.data.player-1].sum_strength += ships[s]->data.compensation_rating;
        }
        fl.status.alive = true;
    }

    /* The following is required for FLAK_ALGORITHM >= 20060624 only,
       but it doesn't hurt to fill it in */
    player_count = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i) {
        if (player_status[i-1].num_live_ships != 0)
            player_index[player_count++] = i;
    }

    for (unsigned i = 0; i < fleets.size(); ++i)
        chooseEnemy(*fleets[i]);
}

/** Get unit by Id.
    \param id Id number
    \param want_planet True if we want a planet, false if we want a ship
    \return requested unit; 0 if not found */
FlakShip*
FlakBattle::getShipById(int id, bool want_planet)
{
    for (ship_t i = 0; i < ships.size(); ++i)
        if (ships[i]->data.id == id && ships[i]->isPlanet() == want_planet)
            return ships[i];
    return 0;
}



static void
writeUnit(std::ostream& os, FlakShip* p)
{
    if (!p)
        os << "none";
    else
        os << "#" << p->data.id;
}

/** Print a checkpoint. This function is used for debugging. It writes
    out almost all state variables to the specified stream. A similar
    checkpoint function is available in the DOS FLAK player; we can
    thus generate traces of a fight which must be identical. */
void
FlakBattle::printCheckpoint(std::ostream& os)
{
    os << "=== " << time << " (seed=" << int32(seed) << ") ===\n";
    for (fleet_t i = 0; i < fleets.size(); ++i) {
        FlakFleet& f = *fleets[i];
        os << "fleet " << i << ": pos=" << f.status.position.x << "," << f.status.position.y
           << " alive="
           << (f.isAlive() ? "TRUE" : "FALSE")
           << " enemy=";
        writeUnit(os, f.status.enemy_ptr);
        os << "\n";
    }
    for (ship_t i = 0; i < ships.size(); ++i) {
        FlakShip& s = *ships[i];
        os << "ship " << i << ": crew=" << int(s.status.crew)
           << " damage=" << int(s.status.damage)
           << " shield=" << int(s.status.shield)
           << " alive=" << (s.isAlive() ? "TRUE": "FALSE")
           << " bs=";
        for (int j = 0; j < 5; ++j)
            os << s.status.beam_charge[j] << " ";
        os << "ts=";
        for (int j = 0; j < 5; ++j)
            os << s.status.torp_charge[j] << " ";
        os << "fs=";
        for (int j = 0; j < 5; ++j)
            os << s.status.bay_charge[j] << " ";
        os << "ftrs_out=" << s.status.fighters_launched
           << " ftrs_in=" << s.status.fighter_count
           << " cdn=" << s.status.launch_countdown
           << " torps=" << s.status.torp_count
           << "\n";
    }
    for (int i = 1; i <= 11; ++i) {
        FlakPlayer& pl = player_status[i-1];
        os << "player " << i << ": num_live_ships="
           << pl.num_live_ships << "\n";
        for (FlakObject* o = pl.stuff; o != 0; o = o->next) {
            os << "  " << int(o->kind) << " at "
               << o->position.x << "," << o->position.y << " ene=";
            writeUnit(os, o->enemy_ptr);
            os << " owner=";
            writeUnit(os, o->owner_ptr);
            os << " kill=" << o->kill << " expl=" << o->expl
               << " speed=" << o->speed << " strikes=" << o->strikes
               << " can_change=" << (o->can_change_enemy ? "TRUE":"FALSE")
               << "\n";
        }
    }
}

/** Play one tick.
    \return true if battle ends. */
bool
FlakBattle::playOneTick()
{
    // recharge
    for (unsigned i = 0; i < ships.size(); ++i)
        ships[i]->recharge();

    // choose enemy
    if (time != 0 && time % FLAK_CHOOSE_ENEMY_TIME == 0)
        for (unsigned i = 0; i < fleets.size(); ++i)
            chooseEnemy(*fleets[i]);

    // launch fighters
    for (unsigned i = 0; i < fleets.size(); ++i)
        launchFighters(*fleets[i]);

    // fire torps
    for (unsigned i = 0; i < fleets.size(); ++i)
        fireTorps(*fleets[i]);

    // fire beams. We choose a random fleet to fire first to achieve
    // random distribution of hits.
    unsigned fleet_off = random(fleets.size());
    for (unsigned i = fleet_off; i < fleets.size(); ++i)
        fireBeams(*fleets[i]);
    for (unsigned i = 0; i < fleet_off; ++i)
        fireBeams(*fleets[i]);

    // fighters fire. We choose a random player to fire first to
    // achieve random distribution of hits
#if FLAK_ALGORITHM >= 20060624
    // Distribution should be uniform among players taking part,
    // not among all players
    unsigned player_off = player_index[random(player_count)] - 1;
#else
    unsigned player_off = random(countof(player_status));
#endif
    for (unsigned i = player_off; i < countof(player_status); ++i)
        fightersFire(player_status[i]);
    for (unsigned i = 0; i < player_off; ++i)
        fightersFire(player_status[i]);

    // fighter intercept
    for (unsigned i = 0; i < countof(player_status)-1; ++i) {
        if (player_status[i].have_any_fighters) {
            for (unsigned j = i+1; j < countof(player_status); ++j) {
                if (player_status[j].have_any_fighters) {
#if FLAK_ALGORITHM >= 20060624
                    // fighterIntercept prefers having a "young" fighter of
                    // the first player firing at an "old" one of the second
                    // one. Since no way is known to fix that, we randomly
                    // swap the players' roles.
                    if (random(2) == 0)
                        fighterIntercept(player_status[i], player_status[j]);
                    else
                        fighterIntercept(player_status[j], player_status[i]);
#else
                    fighterIntercept(player_status[i], player_status[j]);
#endif
                }
            }
        }
    }

    // move stuff
    for (unsigned i = player_off; i < countof(player_status); ++i)
        moveStuff(player_status[i]);
    for (unsigned i = 0; i < player_off; ++i)
        moveStuff(player_status[i]);

    // gc
    for (unsigned i = 0; i < fleets.size(); ++i)
        doFleetGC(*fleets[i]);

    // playergc
    for (unsigned i = 0; i < countof(player_status); ++i)
        player_status[i].doPlayerGC();

    // move units
    for (unsigned i = 0; i < fleets.size(); ++i)
        computeNewPosition(*fleets[i]);
    for (unsigned i = 0; i < fleets.size(); ++i)
        fleets[i]->status.position = fleets[i]->new_position;

    //printCheckpoint(std::cout);

    ++time;

    // end check
    is_terminated = endCheck();
    return is_terminated;
}

/**************************** FlakBattle Setup ***************************/

/** Construct empty battle object. */
FlakBattle::FlakBattle()
    : seed(0), time(0), orig_seed(0), is_terminated(false), vis(0)
{
    for (ship_t i = 0; i < countof(player_status); ++i)
        player_status[i].init(i+1);
}

/** Copy constructor. */
FlakBattle::FlakBattle(const FlakBattle& b)
    : attack_list(b.attack_list),
      seed(b.seed), time(b.time), orig_seed(b.orig_seed), is_terminated(b.is_terminated), vis(0)
{
    /* copy players */
    for (ship_t i = 0; i < countof(player_status); ++i) {
        player_status[i].init(i+1);
        player_status[i] = b.player_status[i];
    }

    /* copy fleets */
    for (fleet_t i = 0; i < b.getNumFleets(); ++i)
        safeAppend(fleets, new FlakFleet(b.getFleetByNumber(i)));

    /* copy ships */
    for (ship_t i = 0; i < b.getNumShips(); ++i)
        safeAppend(ships, new FlakShip(b.getShipByNumber(i)));

    /* enter ship links */
    for (fleet_t i = 0; i < getNumFleets(); ++i) {
        FlakFleet& me = getFleetByNumber(i);
        for (ship_t s = 0; s < me.data.num_ships; ++s) {
            FlakShip& sh = getShipByNumber(me.data.first_ship + s);
            sh.fleet_link = &me;
            sh.battle_link = this;
        }
    }
}

/** Destructor. */
FlakBattle::~FlakBattle()
{
    safeClear(fleets);
    safeClear(ships);
}

/** Add an empty fleet for the given player. You can then add ships to
    the fleet using addShip(). */
FlakBattle::fleet_t
FlakBattle::addFleet(int player)
{
    TFlakFleet data;
    data.player     = player;
    data.first_ship = ships.size();
    data.num_ships  = 0;
    data.speed      = 0;
    data.x_init     = 0;
    data.y_init     = 0;
    safeAppend(fleets, new FlakFleet(data));
    return fleets.size()-1;
}

/** Add ship to the combat. The ship will be added to the most-recently added
    fleet.
    \pre there already is a fleet which matches the ship's owner
    \return number assigned to this ship (running index) */
FlakBattle::ship_t
FlakBattle::addShip(const TFlakShip& ship)
{
    ASSERT(time == 0);
    ASSERT(!fleets.empty());
    ASSERT(ship.player == fleets.back()->data.player);

    FlakFleet& the_fleet = *fleets.back();
    ASSERT(int(ships.size()) == the_fleet.data.first_ship + the_fleet.data.num_ships);

    the_fleet.data.num_ships++;

    safeAppend(ships, new FlakShip(&the_fleet, this, ship));
    return ships.size()-1;
}

/** Add entry to attack list. This only populates the attack list, it
    does not assign the entry to a particular fleet.
    \param ship_t ship to attack
    \param rating_bonus rating bonus awarded for this attack
    \return attack list slot used for this entry (running index) */
int32
FlakBattle::addAttackListEntry(int16 ship_id, int16 rating_bonus)
{
    ASSERT(time == 0);
    attack_list.push_back(ship_id);
    attack_list.push_back(rating_bonus);
    return attack_list.size() / 2 - 1;
}

/** Set initial seed. */
void
FlakBattle::setInitialSeed(uint32 seed)
{
    ASSERT(time == 0);
    this->seed = this->orig_seed = seed;
}

/** Encode this battle into byte block.
    \param s             target buffer
    \param x,y           position, game universe coordinates
    \param ambient_flags ambient flags to place in data block
    \pre battle already run, and placed at end */
void
FlakBattle::saveToStream(Buffer<char>& s, int x, int y, int32 ambient_flags)
{
    ASSERT(time != 0);
    ASSERT(attack_list.size() % 2 == 0);
    ASSERT(is_terminated);

    TFlakBattle data;
    data.x                    = x;
    data.y                    = y;
    data.seed                 = orig_seed;
    data.num_fleets           = fleets.size();
    data.num_ships            = ships.size();
    data.num_att_list_entries = attack_list.size() / 2;
    data.total_time           = time;
    data.ambient_flags        = ambient_flags;
    data.fleet_entry_size     = TFlakFleet::size;
    data.fleet_ptr            = TFlakBattle::size;
    data.ship_entry_size      = TFlakShip::size;
    data.ship_ptr             = data.fleet_ptr + TFlakFleet::size * fleets.size();
    data.att_list_entry_size  = 4;
    data.att_list_ptr         = data.ship_ptr + TFlakShip::size * ships.size();
    data.this_size            = data.att_list_ptr + 4 * data.num_att_list_entries;

    storeAppendStructure(s, data);
    for (fleet_t i = 0; i < fleets.size(); ++i)
        storeAppendStructure(s, fleets[i]->data);
    for (ship_t i = 0; i < ships.size(); ++i)
        storeAppendStructure(s, ships[i]->data);

    /* attack list */
    std::size_t index = s.getSize();
    s.ensureSize(index + 2*attack_list.size());
    for (std::size_t i = 0; i < attack_list.size(); ++i) {
        storeInt16(&s[index], attack_list[i]);
        index += 2;
    }
}

/** Load battle from stream.
    \param s          stream to read
    \param header_out [out] header will be returned here
    \pre object is empty
    \throw FileFormatException on error */
void
FlakBattle::loadFromStream(string_t name, const Buffer<char>& s, TFlakBattle& header_out)
{
    ASSERT(time == 0);
    ASSERT(ships.size() == 0);
    ASSERT(fleets.size() == 0);
    ASSERT(attack_list.size() == 0);
    ASSERT(!is_terminated);

    getStructure(&s[0], header_out);

    if (header_out.fleet_entry_size != TFlakFleet::size || header_out.ship_entry_size != TFlakShip::size || header_out.att_list_entry_size != 2*sizeof(int16))
        throw FileFormatException(name, _("Unsupported file format"));
    if (!header_out.num_fleets || !header_out.num_ships || !header_out.num_att_list_entries)
        throw FileFormatException(name, _("Invalid file format (object count is zero)"));

    setInitialSeed(header_out.seed);

    /* load fleets */
    std::size_t ptr = header_out.fleet_ptr;
    int32 ship_counter = 0;
    for (int32 i = 0; i < header_out.num_fleets; ++i) {
        TFlakFleet f;
        getStructure(&s[ptr], f);

        if (f.first_ship != ship_counter || f.first_ship + f.num_ships > header_out.num_ships)
            throw FileFormatException(name, _("Invalid file format (inconsistent ship numbering)"));
        if (f.att_list_pointer < 0 || f.att_list_pointer > header_out.num_att_list_entries
            || f.att_list_size < 0 || f.att_list_pointer + f.att_list_size > header_out.num_att_list_entries)
            throw FileFormatException(name, _("Invalid file format (invalid attack list pointer)"));

        ship_counter += f.num_ships;
        safeAppend(fleets, new FlakFleet(f));
        ptr += header_out.fleet_entry_size;
    }
    if (ship_counter != header_out.num_ships)
        throw FileFormatException(name, _("Invalid file format (inconsistent ship numbering)"));

    /* load ships */
    ptr = header_out.ship_ptr;
    std::size_t fleet_counter = 0;
    for (int32 i = 0; i < header_out.num_ships; ++i) {
        TFlakShip sh;
        getStructure(&s[ptr], sh);

        if (sh.player != fleets[fleet_counter]->data.player)
            throw FileFormatException(name, _("Invalid file format (invalid owner information)"));

        safeAppend(ships, new FlakShip(fleets[fleet_counter], this, sh));
        if (ships.size() >= fleets[fleet_counter]->data.first_ship + fleets[fleet_counter]->data.num_ships)
            ++fleet_counter;

        ptr += header_out.ship_entry_size;
    }

    /* attack list */
    ptr = header_out.att_list_ptr;
    for (int32 i = 0; i < header_out.num_att_list_entries; ++i) {
        attack_list.push_back(getInt16(&s[ptr]));
        attack_list.push_back(getInt16(&s[ptr+2]));
        ptr += header_out.att_list_entry_size;
    }
}
