/**
  *  \file game/prefio.cc
  *  \brief User Preferences I/O
  *
  *  FIXME: having this in a separate module means that checkinits.pl
  *  will not warn of a missing mod_pref.
  *
  *  Up to beta 12, getUserSettingsDirectory() was implemented to
  *  return $HOME resp. %APPDATA%. Since beta 13, it is a subdirectory
  *  of said folder (.pcc2 under Unix, PCC2 under Windows). Right now,
  *  we load the file from both locations, and save it to where we got
  *  it from. This means we keep interchangability with previous
  *  versions, in case a user doesn't like this version and goes back
  *  to the previous one.
  *
  *  \todo A future version should migrate the pcc2.ini file into the
  *  subdirectory.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/pref.h"
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "io/dirs.h"
#include "io/memstream.h"
#include "io/streams.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/except.h"
#include "game/gamedirs.h"

static int loaded_pref_version;

static const char PCC2_INI[] = "pcc2.ini";

static string_t
getPrefFileName(int version)
{
    return Directory::makeFileName(Platform::getUserSettingsDirectory(version), PCC2_INI);
}

static void
createPath(string_t name)
{
    string_t parent = Directory::getDirName(name);
    string_t child  = Directory::getFileName(name);

    try {
        Ptr<Directory> d = Directory::openDir(parent);
        Ptr<DirectoryEntry> e = d->getEntryByName(child);
        if (e->getFileType() != DirectoryEntry::FT_DIRECTORY)
            e->createAsDirectory();
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
    }
}

/** Create directory name for preference files. */
void
createUserPreferencesDirectory()
{
    createPath(Platform::getUserSettingsDirectory(1));
}

/** Load user preferences file (pcc2.ini). */
static void
loadUserPreferences()
{
    int version = 1;
    Ptr<Stream> s = Stream::openFileNT(getPrefFileName(version), Stream::C_READ);
    if (!s) {
        version = 0;
        s = Stream::openFileNT(getPrefFileName(version), Stream::C_READ);
    }
    if (s) {
        // Parse the file
        console.write(LOG_DEBUG, format(_("Loading %s..."), s->getName()));
        ConfigFileParser cfp(getUserPreferences(), true, ConfigOption::srcUser);
        cfp.setUtf8();
        cfp.parseFile(*s);

        // Set all options to srcUser, no matter where they come from.
        // This will make sure the main config file always contains all options.
        for (Config::iterator i = getUserPreferences().begin(); i != getUserPreferences().end(); ++i) {
            i->setSource(ConfigOption::srcUser);
        }

        // Remember preferences version to write the file in the same place
        loaded_pref_version = version;
    } else {
        // No file found, so create it with the new version.
        loaded_pref_version = 1;
    }
}

/** Load directory preferences file on top of user preferences. */
static void
loadDirectoryPreferences()
{
    Ptr<Stream> s = game_file_dir->openFileNT(PCC2_INI, Stream::C_READ);
    if (s) {
        console.write(LOG_DEBUG, format(_("Loading %s..."), s->getName()));
        ConfigFileParser cfp(getUserPreferences(), true, ConfigOption::srcGame);
        cfp.setUtf8();
        cfp.parseFile(*s);
    }
}

/** Save user preferences file (pcc2.ini). */
static void
saveUserPreferences()
{
    string_t name = getPrefFileName(loaded_pref_version);

    // Update config file in memory
    MemoryStream ms;
    ConfigFileUpdater cfu(getUserPreferences(), ms, ConfigOption::srcUser);
    cfu.setUtf8();
    if (Ptr<Stream> s = Stream::openFileNT(name, Stream::C_READ)) {
        cfu.parseFile(*s);
    }
    cfu.finish();

    // Create path
    if (loaded_pref_version > 0) {
        createUserPreferencesDirectory();
    }

    // Write it out
    try {
        Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE | Stream::C_WRITE);
        console.write(LOG_DEBUG, format(_("Writing %s..."), s->getName()));
        s->writeT(ms.getBuffer().ptr(), ms.getBuffer().getSize());
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
    }
}

/** Save all preferences, and reload user preferences. */
static void
savePreferencesAndReloadUser()
{
    // Save user preferences first. If the user challenges his fate by entering
    // the config file directory as a game, all options will be marked srcGame.
    // saveUserPreferences() will therefore create a blank file. By saving
    // directory preferences after that, we'll rewrite the file with content.
    saveUserPreferences();

    // Update directory preferences
    MemoryStream ms;
    ConfigFileUpdater cfu(getUserPreferences(), ms, ConfigOption::srcGame);
    cfu.setUtf8();
    if (Ptr<Stream> s = game_file_dir->openFileNT(PCC2_INI, Stream::C_READ)) {
        cfu.parseFile(*s);
    }
    cfu.finish();

    // Save it
    if (ms.getSize() == 0) {
        // File would be empty, so delete it
        game_file_dir->eraseNT(PCC2_INI);
    } else {
        try {
            Ptr<Stream> s = game_file_dir->openFile(PCC2_INI, Stream::C_CREATE | Stream::C_WRITE);
            console.write(LOG_DEBUG, format(_("Writing %s..."), s->getName()));
            s->writeT(ms.getBuffer().ptr(), ms.getBuffer().getSize());
        }
        catch (FileProblemException& e) {
            console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
        }
    }

    // Reload user preferences. This will remove all per-game settings.
    // Since we (try to) keep all options in the per-user file, this
    // will restore the global settings.
    loadUserPreferences();
}

extern const ModuleInitializer mod_pref[] = {
    { OnStartup,                   loadUserPreferences },
    { OnEnterDirectory,            loadDirectoryPreferences },
    { OnLeftDirectory,             savePreferencesAndReloadUser },
    { OnShutdown,                  saveUserPreferences },
    // { OnEnterNoGame | OnEnterDirectory, loadUserPreferences },
    // { OnLeftDirectory,                  saveUserPreferences },
    { 0, 0 },
};
