/**
  *  \file int/file.cc
  *  \brief Interpreter: File I/O Related Stuff
  *
  *  This module implements all file I/O related functions. Some are
  *  implemented as builtin commands due to their special syntax,
  *  which internally call regular commands.
  *
  *  \todo This generates and evaluates <tt>ufilenr</tt> instructions.
  *  Those are actually not required at runtime; we know we're dealing
  *  with file numbers. We can, however, not remove them completely at
  *  compile time because many commands are called regularily, and the
  *  compiler makes no further assumptions about regular commands.
  *  Removing the <tt>ufilenr</tt> instructions completely would mean
  *  we accept file numbers in invalid places.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "int/file.h"
#include "int/special.h"
#include "cpluslib/ptrvector.h"
#include "int/interror.h"
#include "int/expr.h"           // IntExprNode
#include "int/lex.h"            // IntTokenizer
#include "int/statement.h"      // parseArgumentList
#include "int/args.h"
#include "int/if/ifutil.h"      // FIXME
#include "int/compound.h"
#include "int/intglobal.h"      // defineGlobalValue
#include "io/io.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "int/parseexpr.h"
#include "util/recode.h"
#include "io/dirs.h"
#include "util/console.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "util/nls.h"

/** Size limit for blobs. This is an artifical limit to avoid that
    errors overload the program too easily. */
enum { BLOB_LIMIT = 65536 };

/** Check Blob-type argument.
    \param out [out] Blob returned here
    \param p [in] Argument received from user
    \return true iff valid blob found, false if user passed null
    \throw IntError on type error */
static bool
checkBlobArg(IntBlobValue*& out, IntValue* p)
{
    if (!p) {
        return false;
    } else if (IntBlobValue* bv = dynamic_cast<IntBlobValue*>(p)) {
        out = bv;
        return true;
    } else {
        throw IntError::typeError(IntError::tyBlob);
    }
}

/**************************** File Descriptors ***************************/

/** Maximum number of user files. */
enum { MAX_SCRIPT_FILES = 101 };

/** Open modes. We use those internally instead of Stream::C_FOO,
    because the numerical values of C_FOO are an implementation detail
    to class Stream, but those values should remain portable across
    saved VMs. */
enum {
    OpenForReading,
    OpenForWriting,
    OpenForRandom,
    OpenForAppend,
    OpenModeMax = OpenForAppend
};

namespace {
    struct IntFileDescriptor {
        Ptr<Stream> stream;
        TextFile text_file;

        IntFileDescriptor(Ptr<Stream> s);
        ~IntFileDescriptor();
    };
}

IntFileDescriptor::IntFileDescriptor(Ptr<Stream> s)
    : stream(s),
      text_file(*s)
{
    text_file.setCharacterSet(getGameCharacterSet());
}

IntFileDescriptor::~IntFileDescriptor()
{
    try {
        text_file.sync();
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format("%s: %s", e.getFileName(), e.what()));
        console.write(LOG_ERROR, format(_("%s: Error closing this file, possible data loss."), text_file.getName()));
    }
}

/** Open script files. Note that slot 0 exists here, but we never
    allocate it to the user unless he explicitly requests it. It is
    reserved to the implementation, just in case. In PCC 1.x, file #0
    was reserved to the core, and not accessible through the script
    side. It was used to implement the user interface's selection I/O
    functions in terms of the script SaveSelection/LoadSelection
    implementation. */
static ptr_vector<IntFileDescriptor> script_fds;

/** Check file number argument.
    \param fd [out] File number, guaranteed to be in range
    \param arg [in] Argument received from user
    \param mustBeOpen [in] If true, accept only open files
    \return true iff argument was given, false if it was null
    \throw IntError on type error */
static bool
checkFileArg(int& fd, IntValue* arg, bool mustBeOpen)
{
    // Check for null
    if (!arg)
        return false;

    // Check for file number
    int32_t value;
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg))
        value = iv->getValue();
    else if (IntFileValue* fv = dynamic_cast<IntFileValue*>(arg))
        value = fv->getFileNumber();
    else
        throw IntError::typeError(IntError::tyFile);

    // Check range
    if (value < 0 || value >= MAX_SCRIPT_FILES)
        throw IntError::rangeError();
    if (mustBeOpen && (value >= int(script_fds.size()) || script_fds[value] == 0))
        throw IntError("File not open");

    fd = value;
    return true;
}

/** Check file number argument, public interface.
    \param tf [out] File handle, guaranteed to be valid if return value is true
    \param arg [in] Argument received from user
    \return true iff argument was given, false if it was null
    \throw IntError on type error */
bool
checkFileArg(TextFile*& tf, IntValue* arg)
{
    int fd;
    if (checkFileArg(fd, arg, true)) {
        tf = &script_fds[fd]->text_file;
        return true;
    } else {
        tf = 0;
        return false;
    }
}

/** Open a new file.
    \param fd  File number
    \param sfd Newly-allocated IntFileDescriptor object */
static void
openNewFile(int fd, IntFileDescriptor* sfd)
{
    try {
        while (script_fds.size() <= uint32_t(fd))
            script_fds.push_back(0);
        script_fds.replace_elem(fd, sfd);
    }
    catch (...) {
        delete sfd;
        throw;
    }
}

/****************************** Compilation ******************************/

/** Compile a "SetXxx" command. The command is compiled into a
    read-modify-write cycle, using the CC$SETINT (IFCCSetInt)
    function. Parameters given by the user are the blob variable,
    an index, and the values.
    \param tok [in] Tokenizer to read user parameters from
    \param bco [in/out] Byte-code object to produce output on
    \param scc [in] Statement compilation context
    \param size [in] Number of bytes to pack each parameter in */
static void
compileSet(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc, const int size)
{
    // Read arguments
    ptr_vector<IntExprNode> xn;
    tok.readNext();
    parseCommandArgumentList(tok, xn);
    if (xn.size() < 3)
        throw IntError("Too few arguments");

    // Read cycle for first arg
    xn[0]->compileRead(bco, scc);

    // Push size
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sInteger, size);

    // Compile all other args
    for (uint32_t i = 1; i < xn.size(); ++i)
        xn[i]->compileValue(bco, scc);

    // Call routine to do the work
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("CC$SETINT"));
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad, xn.size() + 1);

    // Write cycle for first arg
    xn[0]->compileWrite(bco, scc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
}

/* @q SetByte v:Blob, pos:Int, value:Int... (Global Command)
   Store bytes into blob.
   Packs the %value arguments into the blob %v, one byte per element,
   starting at position %pos.
   The first position in the blob has index 0.
   @since PCC2 1.99.12, PCC 1.0.13 */
static void
SFSetByte(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    compileSet(tok, bco, scc, 1);
}

/* @q SetWord v:Blob, pos:Int, value:Int... (Global Command)
   Store words into blob.
   Packs the %value arguments into the blob %v, two bytes (16 bits) per element,
   starting at position %pos.
   The first position in the blob has index 0.
   @since PCC2 1.99.12, PCC 1.0.13 */
static void
SFSetWord(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    compileSet(tok, bco, scc, 2);
}

/* @q SetLong v:Blob, pos:Int, value:Int... (Global Command)
   Store longs into blob.
   Packs the %value arguments into the blob %v, four bytes (32 bits) per element,
   starting at position %pos.
   The first position in the blob has index 0.
   @since PCC2 1.99.12, PCC 1.0.13 */
static void
SFSetLong(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    compileSet(tok, bco, scc, 4);
}

/* @q SetStr v:Blob, pos:Int, length:Int, str:Int (Global Command)
   Store string into blob.
   The string is converted to the game character set, padded with spaces or truncated to match the %length,
   and then stored into the blob %v starting at position %pos.
   The first position in the blob has index 0.
   @since PCC2 1.99.12, PCC 1.0.13 */

/** Compile a "SetStr" command. This command packs a string into a
    space-padded, game-charset string. It is compiled into a
    read-modify-write cycle using the CC$SETSTR (IFCCSetStr) function.
    Arguments given by the user are
    - data block (blob)
    - index
    - length of field
    - value */
static void
SFSetStr(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    // Read arguments
    ptr_vector<IntExprNode> xn;
    tok.readNext();
    parseCommandArgumentList(tok, xn);
    checkArgumentCount(xn.size(), 4, 4);

    // Read cycle for first arg
    xn[0]->compileRead(bco, scc);

    // Compile all other args
    for (uint32_t i = 1; i < xn.size(); ++i)
        xn[i]->compileValue(bco, scc);

    // Call routine to do the work
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("CC$SETSTR"));
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad, xn.size());

    // Write cycle for first arg
    xn[0]->compileWrite(bco, scc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
}

/* @q Open name:Str For Input|Output|Random|Append As #file:File] (Global Command)
   Open a file.
   The %mode specifies what you intend to do with the file:
   - %Input: open existing file for reading
   - %Output: create new file for writing, overwrite existing file
   - %Random: open existing file for reading and writing
   - %Append: append to existing file or create new one

   Operations with this file will use <a href="int:index:type:file">file number</a> %file.
   If that number already referred to an existing file before, that one will be closed first.
   @since PCC2 1.99.12, PCC 1.0.13 */

/** Compile "Open" command. Syntax is
    "Open FILENAME (For (Input|Output|Random|Append)|As #fd)*".
    This is compiled into "CC$Open FD, FILENAME, MODE". */
static void
SFOpen(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    // Read file name argument
    tok.readNext();
    std::auto_ptr<IntExprNode> file_name(IntExpressionParser(tok).parse());

    // Read keyword arguments
    std::auto_ptr<IntExprNode> fd;
    int mode = -1;
    while (tok.getCurrentToken() != tok.tEnd) {
        if (tok.checkAdvance("FOR")) {
            if (mode != -1)
                throw IntError("Duplicate mode for 'Open'");
            if (tok.checkAdvance("INPUT"))
                mode = OpenForReading;
            else if (tok.checkAdvance("OUTPUT"))
                mode = OpenForWriting;
            else if (tok.checkAdvance("RANDOM"))
                mode = OpenForRandom;
            else if (tok.checkAdvance("APPEND"))
                mode = OpenForAppend;
            else
                throw IntError("Invalid mode for 'Open'");
        } else if (tok.checkAdvance("AS")) {
            if (fd.get())
                throw IntError("Duplicate file number for 'Open'");
            fd.reset(IntExpressionParser(tok).parse());
        } else {
            throw IntError("Syntax error");
        }
    }

    // Do we have everything?
    if (mode == -1)
        throw IntError("Missing mode for 'Open'");
    if (!fd.get())
        throw IntError("Missing file number for 'Open'");

    // Generate code
    fd->compileValue(bco, scc);
    file_name->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sInteger, mode);
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("CC$OPEN"));
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 3);
}

/* @q Get #f:File, var:Blob, length:Int (Global Command)
   Read binary data.
   This command will read %length bytes from the file, and place them in the data block %var.
   @since PCC2 1.99.12, PCC 1.0.13 */

/** Compile "Get" command. Syntax is "Get #1, data, length". This is
    compiled into "data := CC$Get(data, #1, length)". */
static void
SFGet(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    // Read arguments
    ptr_vector<IntExprNode> xn;
    tok.readNext();
    parseCommandArgumentList(tok, xn);
    checkArgumentCount(xn.size(), 3, 3);

    // Generate code. We must generate a read-modify-write cycle, because if
    // the fd or the length is null, data must remain unchanged.
    xn[1]->compileRead(bco, scc);
    xn[0]->compileValue(bco, scc);
    xn[2]->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("CC$GET"));
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad, 3);
    xn[1]->compileWrite(bco, scc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
}

/* @q Input #f:File, var:Str, Optional flag:Bool (Global Command)
   Read line data.
   This command will read one line of text from the file, and place it in the variable %var.
   The %flag specifies what to do at the end of the file:
   - %False (default): generate an error
   - %True: set %var to EMPTY and continue normally
   @since PCC2 1.99.12, PCC 1.0.13 */

/** Compile "Input" command. Syntax is "Input #1, output, flag". This is
    compiled into "data := CC$Input(data, #1[, flag])". */
static void
SFInput(IntTokenizer& tok, IntBytecodeObject& bco, const IntStatementCompilationContext& scc)
{
    // Read arguments
    ptr_vector<IntExprNode> xn;
    tok.readNext();
    parseCommandArgumentList(tok, xn);
    checkArgumentCount(xn.size(), 2, 3);

    // Generate code. We must generate a read-modify-write cycle, because if
    // the fd or the length is null, data must remain unchanged.
    xn[1]->compileRead(bco, scc);
    xn[0]->compileValue(bco, scc);
    if (xn.size() > 2)
        xn[2]->compileValue(bco, scc);
    bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("CC$INPUT"));
    bco.addInstruction(IntOpcode::maIndirect, IntOpcode::miIMLoad, xn.size());
    xn[1]->compileWrite(bco, scc);
    bco.addInstruction(IntOpcode::maStack, IntOpcode::miStackDrop, 1);
}

/******************************* Execution *******************************/

/** Prepare blob value.
    \param newBV [out] Newly-allocated, empty blob value
    \param v     [in] Old value
    \param bytes [in] Minimum number of bytes the new blob must have */
static void
prepareBlob(IntBlobValue& newBV, IntValue* v, uint32_t bytes)
{
    std::vector<char>& newValue = newBV.getValue();
    if (IntBlobValue* oldBV = dynamic_cast<IntBlobValue*>(v)) {
        newValue = oldBV->getValue();
    }
    if (newValue.size() < bytes) {
        newValue.insert(newValue.end(), bytes - newValue.size(), char(0));
    }
}

/* @q CC$SetInt(v:Blob, size:Int, pos:Int, value:Int...):Blob (Internal)
   Backend to {SetByte}, {SetWord}, {SetLong}. */
static IntValue*
IFCCSetInt(IntArgBlock& args)
{
    /* Implementation notes: PCC 1.x does not modify the output value
       if one of the parameters is null. This means we have to return
       a copy of our first argument for the same effect.

       In addition, PCC 1.x blanks the blob (a string, actually) if it
       has the wrong type. So let's do the same. It's quite convenient
       to implement anyway, and keeps code which resets the blob using
       things like "blob := ''" working. */
    int32_t size, index;

    // At least four args:
    // - blob
    // - size
    // - index
    // - one or more data elements
    args.checkArgumentCountAtLeast(4);

    IntValue* first = args.getNext();
    if (!checkIntArg(size, args.getNext(), 1, 4) || !checkIntArg(index, args.getNext(), 0, BLOB_LIMIT)) {
        // size or index are null, ignore this command
        return IntValue::cloneOf(first);
    }

    // prepare the blob
    uint32_t bytesNeeded = args.getNumArgs() * size;
    std::auto_ptr<IntBlobValue> blob(new IntBlobValue());
    prepareBlob(*blob, first, index + bytesNeeded);

    // execute
    while (args.getNumArgs() != 0) {
        // read argument
        int32_t theValue;
        if (!checkIntArg(theValue, args.getNext())) {
            return IntValue::cloneOf(first);
        }
        // encode it
        for (int32_t i = 0; i < size; ++i) {
            blob->getValue()[index] = char(theValue & 255);
            ++index;
            theValue >>= 8;
        }
    }

    return blob.release();
}

/* @q CC$SetStr(v:Blob, pos:Int, size:Int, value:Str):Blob (Internal)
   Backend to {SetStr}. */
static IntValue*
IFCCSetStr(IntArgBlock& args)
{
    int32_t index, size;
    string_t value;

    // Four args:
    // - blob
    // - index
    // - size
    // - string
    args.checkArgumentCount(4);

    IntValue* first = args.getNext();
    if (!checkIntArg(index, args.getNext(), 0, BLOB_LIMIT)
        || !checkIntArg(size, args.getNext(), 0, BLOB_LIMIT)
        || !checkStringArg(value, args.getNext()))
    {
        // some parameter is null, ignore this command
        return IntValue::cloneOf(first);
    }

    // Convert to game character set
    value = convertUtf8ToGame(value);

    // prepare the blob
    std::auto_ptr<IntBlobValue> blob(new IntBlobValue());
    prepareBlob(*blob, first, index + size);

    // execute
    if (size != 0)
        storeBasicStringN(&blob->getValue()[index], size, value);

    return blob.release();
}

/** Execute GetByte/Word/Long function.
    \param args User-supplied args: blob, index
    \param size Number of bytes per integer, 1, 2 or 4.
    \return Return value for user, newly-allocated IntValue object */
static IntValue*
extractInt(IntArgBlock& args, int size)
{
    IntBlobValue* bv;
    int32_t index;

    // Read args
    args.checkArgumentCount(2);
    if (!checkBlobArg(bv, args.getNext()) || !checkIntArg(index, args.getNext(), 0, BLOB_LIMIT))
        return 0;

    // Do it
    uint32_t result = 0;
    for (int i = 0; i < size && uint32_t(index) < bv->getValue().size(); ++i) {
        result += uint32_t(uint8_t(bv->getValue()[index])) << (8*i);
        ++index;
    }

    // Byte results are unsigned. For signed long results, conversion
    // uint32_t->int32_t does the right thing. We just have to fix up
    // signed word results.
    if (size == 2) {
        return makeIntValue(int16_t(result));
    } else {
        return makeIntValue(result);
    }
}

/* @q GetByte(v:Blob, pos:Int):Int (Function)
   Extract byte.
   Returns the byte stored at position %pos in the given data block.
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFGetByte(IntArgBlock& args)
{
    return extractInt(args, 1);
}

/* @q GetWord(v:Blob, pos:Int):Int (Function)
   Extract word.
   Returns the word (2 bytes, 16 bits) stored at position %pos in the given data block.
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFGetWord(IntArgBlock& args)
{
    return extractInt(args, 2);
}

/* @q GetLong(v:Blob, pos:Int):Int (Function)
   Extract long.
   Returns the long (4 bytes, 32 bits) stored at position %pos in the given data block.
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFGetLong(IntArgBlock& args)
{
    return extractInt(args, 4);
}

/* @q GetStr(v:Blob, pos:Int, length:Int):Str (Function)
   Extract string.
   Returns the string that is stored at position %pos in the data block in a field of size %length.
   The string is converted from the game character set, and trailing space is removed.
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFGetStr(IntArgBlock& args)
{
    IntBlobValue* bv;
    int32_t index, size;

    // Read args
    args.checkArgumentCount(3);
    if (!checkBlobArg(bv, args.getNext()) || !checkIntArg(index, args.getNext(), 0, BLOB_LIMIT) || !checkIntArg(size, args.getNext(), 0, BLOB_LIMIT))
        return 0;

    // Do it
    string_t result;
    if (uint32_t(index) < bv->getValue().size()) {
        int32_t avail = bv->getValue().size() - index;
        if (avail < size)
            size = avail;
        result = convertGameToUtf8(getBasicStringN(&bv->getValue()[index], size));
    }
    return makeStringValue(result);
}

/* @q CC$Get(old:Blob, fd:File, length:Int):Blob (Internal)
   Backend to {Get}. */
static IntValue*
IFCCGet(IntArgBlock& args)
{
    // Read args
    int fd;
    int32_t size;
    args.checkArgumentCount(3);
    IntValue* oldValue = args.getNext();
    if (!checkFileArg(fd, args.getNext(), true) || !checkIntArg(size, args.getNext(), 0, BLOB_LIMIT))
        return IntValue::cloneOf(oldValue);

    // Do it
    std::auto_ptr<IntBlobValue> bv(new IntBlobValue());
    bv->getValue().resize(size);
    if (size != 0) {
        int32_t got = script_fds[fd]->text_file.readBytes(&bv->getValue()[0], size);
        if (got != size)
            throw IntError("Premature end of file");
    }
    return bv.release();
}

/* @q CC$Input(old:Str, fd:File, Optional flag:Bool):Str (Internal)
   Backend to {Input}. */
static IntValue*
IFCCInput(IntArgBlock& args)
{
    // Read args
    int fd;
    bool flag = false;
    args.checkArgumentCount(2, 3);
    IntValue* oldValue = args.getNext();
    if (!checkFileArg(fd, args.getNext(), true))
        return IntValue::cloneOf(oldValue);
    checkBoolArg(flag, args.getNext());

    // Do it
    string_t line;
    if (script_fds[fd]->text_file.getLine(line))
        return makeStringValue(line);
    else if (flag)
        return 0;
    else
        throw IntError("Premature end of file");
}

/* @q CC$Open(fd:File, name:Str, mode:Int):Any (Internal)
   Backend to {Open}. */

/** Execute "Open" command. Parameters are
    - file number
    - file name
    - mode

    Open takes named arguments, which are currently encoded into the
    mode. Future expansions to the Open command will add bits to mode
    ("feature X present"), and additional arguments ("this is the
    value of feature X"). */
static IntValue*
IFCCOpen(IntArgBlock& args)
{
    // Read args
    int fd;
    string_t filename;
    int32_t mode;

    args.checkArgumentCount(3);
    if (!checkFileArg(fd, args.getNext(), false)
        || !checkStringArg(filename, args.getNext())
        || !checkIntArg(mode, args.getNext(), 0, OpenModeMax))
    {
        return 0;
    }

    // Do it
    if (mode == OpenForAppend) {
        // Append is special
        Ptr<Stream> s(Stream::openFileNT(filename, Stream::C_READ | Stream::C_WRITE));
        if (s) {
            // File already exists. Read one line to detect encoding;
            // UTF-8 files will thus keep their encoding.
            string_t tmp;
            std::auto_ptr<IntFileDescriptor> sfd(new IntFileDescriptor(s));
            sfd->text_file.getLine(tmp);
            sfd->text_file.seek(s->getSize());
            openNewFile(fd, sfd.release());
        } else {
            // File does not exist. Try creating it.
            openNewFile(fd, new IntFileDescriptor(Stream::openFile(filename, Stream::C_WRITE | Stream::C_CREATE)));
        }
    } else if (mode == OpenForWriting) {
        // Write only
        openNewFile(fd, new IntFileDescriptor(Stream::openFile(filename, Stream::C_WRITE | Stream::C_CREATE)));
    } else if (mode == OpenForRandom) {
        // Read+Write
        openNewFile(fd, new IntFileDescriptor(Stream::openFile(filename, Stream::C_READ | Stream::C_WRITE)));
    } else {
        // Read only
        openNewFile(fd, new IntFileDescriptor(Stream::openFile(filename, Stream::C_READ)));
    }
    return 0;
}

/* @q CC$Print(fd:File, text:Str):Any (Internal)
   Backend to {Print} to a file. */
static IntValue*
IFCCPrint(IntArgBlock& args)
{
    int fd;
    string_t text;
    args.checkArgumentCount(2);
    if (checkFileArg(fd, args.getNext(), true) && checkStringArg(text, args.getNext()))
        script_fds[fd]->text_file.writeLine(text);
    return 0;
}

/* @q FPos(#fd:File):Int (Function)
   Get current position within a file.
   @see Seek
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFFPos(IntArgBlock& args)
{
    int fd;
    args.checkArgumentCount(1);
    if (!checkFileArg(fd, args.getNext(), true))
        return 0;
    return makeIntValue(script_fds[fd]->text_file.getPos());
}

/* @q FreeFile():Int (Function)
   Get an unused file number.
   If there is no unused file number, fails with an error.
   Note that this function will always return the same value until you {Open} it
   (or {Close} another file).
   It is therefore normally used in the form
   | Dim fd = FreeFile()
   | Open "file" For Input As #fd
   | Dim fd2 = FreeFile()
   | Open "anotherfile" For Input As #fd2
   @since PCC2 1.99.12, PCC 1.0.13 */

/** "FreeFile()" function. Consistently with PCC 1.x, FreeFile never
    returns 0; see the definition of script_fds. */
static IntValue*
IFFreeFile(IntArgBlock& args)
{
    args.checkArgumentCount(0);
    for (uint32_t i = 1; i < MAX_SCRIPT_FILES; ++i)
        if (i > script_fds.size() || script_fds[i] == 0)
            return makeIntValue(i);
    throw IntError("No free file number");
}

/* @q FSize(#fd:File):Int (Function)
   Get size of the file, in bytes.
   @see FPos
   @since PCC2 1.99.12, PCC 1.0.13 */
static IntValue*
IFFSize(IntArgBlock& args)
{
    int fd;
    args.checkArgumentCount(1);
    if (!checkFileArg(fd, args.getNext(), true))
        return 0;
    return makeIntValue(script_fds[fd]->text_file.getSize());
}

/* @q MakeFileName(n:Str...):Str (Function)
   Create a file name.
   Parameters are file name fragments, i.e. directories, until the last fragment which is a file name.
   This function builds a file name from these, using operating-system dependant rules.
   For example,
   | MakeFileName("a", "b", "c.txt")
   will return "a\b\c.txt" or "a/b/c.txt", depending on the operating system.
   @since PCC2 1.99.12 */
static IntValue*
IFMakeFileName(IntArgBlock& args)
{
    string_t a, b;
    args.checkArgumentCountAtLeast(1);
    if (!checkStringArg(a, args.getNext()))
        return 0;

    while (args.getNumArgs() != 0) {
        if (!checkStringArg(b, args.getNext()))
            return 0;
        a = Directory::makeFileName(a, b);
    }
    return makeStringValue(a);
}

/* @q GetFileName(n:Str):Str (Function)
   Get file name.
   The parameter is a full file name, possibly including a directory path.
   This function removes all directory names, and returns just the file name.
   @since PCC2 1.99.12 */
static IntValue*
IFGetFileName(IntArgBlock& args)
{
    string_t a;
    args.checkArgumentCount(1);
    if (!checkStringArg(a, args.getNext()))
        return 0;
    else
        return makeStringValue(Directory::getFileName(a));
}

/* @q GetDirectoryName(n:Str):Str (Function)
   Get directory name.
   The parameter is a full file name, possibly including a directory path.
   This function removes the final file name component and returns just the directories.
   @since PCC2 1.99.12 */
static IntValue*
IFGetDirectoryName(IntArgBlock& args)
{
    string_t a;
    args.checkArgumentCount(1);
    if (!checkStringArg(a, args.getNext()))
        return 0;
    else
        return makeStringValue(Directory::getDirName(a));
}

/* @q Close #fd:File (Global Command)
   Close a file.
   If some data is still in the write buffer, it will be written to disk now.
   The file number will become available for re-use.
   @since PCC2 1.99.12, PCC 1.0.13 */
static void
IFClose(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int fd;
    args.checkArgumentCount(1);
    if (checkFileArg(fd, args.getNext(), true)) {
        script_fds[fd]->text_file.sync();
        script_fds.replace_elem(fd, 0);
    }
}

/* @q Put #fd:File, v:Blob, Optional length:Int (Global Command)
   Write binary data.
   Writes the data block %v into the file at the current position.
   If the %length is specified (recommended), it determines the number of bytes to write.
   If the %length is not specified, PCC writes as many bytes as the block contains.

   @since PCC2 1.99.12, PCC 1.0.13 */
static void
IFPut(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Read args
    int fd;
    IntBlobValue* bv;
    int32_t size;
    args.checkArgumentCount(2, 3);
    if (!checkFileArg(fd, args.getNext(), true) || !checkBlobArg(bv, args.getNext()))
        return;
    if (!checkIntArg(size, args.getNext(), 0, BLOB_LIMIT))
        size = bv->getValue().size();

    // Produce output
    uint32_t usize = size;
    if (usize <= bv->getValue().size()) {
        // Blob has enough data
        script_fds[fd]->text_file.writeBytes(&bv->getValue()[0], usize);
    } else {
        // Blob has not enough data. Write whole blob.
        uint32_t blobSize = bv->getValue().size();
        if (blobSize != 0)
            script_fds[fd]->text_file.writeBytes(&bv->getValue()[0], blobSize);
        usize -= blobSize;

        // Now write some nulls
        while (usize > 0) {
            static const char zeroes[256] = {};
            uint32_t now = usize > sizeof(zeroes) ? sizeof(zeroes) : usize;
            script_fds[fd]->text_file.writeBytes(zeroes, now);
            usize -= now;
        }
    }
}

/* @q Seek #fd:File, pos:Int (Global Command)
   Go to position in file.
   @see FPos(), FSize()
   @since PCC2 1.99.12, PCC 1.0.13 */
static void
IFSeek(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int fd;
    int32_t pos;
    args.checkArgumentCount(2);
    if (checkFileArg(fd, args.getNext(), true) && checkIntArg(pos, args.getNext(), 0, 0x7FFFFFFF)) {
        script_fds[fd]->text_file.seek(pos);
    }
}

void
initInterpreterFileInterface()
{
    defineSpecialCommand("GET",     new IntSimpleSpecialCommand(SFGet));
    defineSpecialCommand("INPUT",   new IntSimpleSpecialCommand(SFInput));
    defineSpecialCommand("OPEN",    new IntSimpleSpecialCommand(SFOpen));
    defineSpecialCommand("SETBYTE", new IntSimpleSpecialCommand(SFSetByte));
    defineSpecialCommand("SETLONG", new IntSimpleSpecialCommand(SFSetLong));
    defineSpecialCommand("SETSTR",  new IntSimpleSpecialCommand(SFSetStr));
    defineSpecialCommand("SETWORD", new IntSimpleSpecialCommand(SFSetWord));
    defineGlobalValue("CC$GET",           new IntSimpleIndexableValue(0, IFCCGet, 0));
    defineGlobalValue("CC$INPUT",         new IntSimpleIndexableValue(0, IFCCInput, 0));
    defineGlobalValue("CC$OPEN",          new IntSimpleIndexableValue(0, IFCCOpen, 0));
    defineGlobalValue("CC$PRINT",         new IntSimpleIndexableValue(0, IFCCPrint, 0));
    defineGlobalValue("CC$SETINT",        new IntSimpleIndexableValue(0, IFCCSetInt, 0));
    defineGlobalValue("CC$SETSTR",        new IntSimpleIndexableValue(0, IFCCSetStr, 0));
    defineGlobalValue("FPOS",             new IntSimpleIndexableValue(0, IFFPos, 0));
    defineGlobalValue("FREEFILE",         new IntSimpleIndexableValue(0, IFFreeFile, 0));
    defineGlobalValue("FSIZE",            new IntSimpleIndexableValue(0, IFFSize, 0));
    defineGlobalValue("GETBYTE",          new IntSimpleIndexableValue(0, IFGetByte, 0));
    defineGlobalValue("GETDIRECTORYNAME", new IntSimpleIndexableValue(0, IFGetDirectoryName, 0));
    defineGlobalValue("GETFILENAME",      new IntSimpleIndexableValue(0, IFGetFileName, 0));
    defineGlobalValue("GETLONG",          new IntSimpleIndexableValue(0, IFGetLong, 0));
    defineGlobalValue("GETSTR",           new IntSimpleIndexableValue(0, IFGetStr, 0));
    defineGlobalValue("GETWORD",          new IntSimpleIndexableValue(0, IFGetWord, 0));
    defineGlobalValue("MAKEFILENAME",     new IntSimpleIndexableValue(0, IFMakeFileName, 0));
    defineGlobalProcedure("CLOSE", IFClose);
    defineGlobalProcedure("PUT",   IFPut);
    defineGlobalProcedure("SEEK",  IFSeek);
}
